from abc import ABCMeta
from dataclasses import dataclass
from io import StringIO
import itertools
import json
import logging
import typing
from typing import Dict, List, Literal, Optional, Tuple, overload
from collections.abc import Callable, Sequence, Iterator, Iterable
from argparse import ArgumentParser, FileType
from importlib import resources
import sys
import warnings
import unicodedata

import panphon
import pandas as pd
from pandas import CategoricalDtype, DataFrame, Index, Series
from pandas.io.parsers.readers import ReadCsvBuffer
from torch import LongTensor, Tensor
import torch
import numpy as np

from allophant.config import Config, PhonemeLayerType, ProjectionEntryConfig
from allophant.package_data import ALLOPHOIBLE_PATH, DEFAULT_DIALECTS_PATH
from allophant.phonemes import MissingSegmentError, IpaSegmenter
from allophant.language_codes import LanguageCode
from allophant.config import ProjectionEntryConfig, FeatureSet
from allophant import utils
from allophant.utils import AnyPath
from allophant.phoneme_segmentation import IpaSentenceSegmenter
from allophant import phoneme_segmentation
from allophant import language_codes


# IPA tie character
TIE = "\u0361"


@dataclass
class LanguageAllophoneMappings:
    allophones: Dict[int, Dict[int, List[int]]]
    languages: List[str]
    shared_phones: List[str]

    def iso6393_inventories(self, shared_phoneme_inventory: Sequence[str]) -> Dict[str, List[str]]:
        return {
            LanguageCode.from_str(language).alpha3: [
                shared_phoneme_inventory[phoneme_index] for phoneme_index in self.allophones[language_id].keys()
            ]
            for language_id, language in enumerate(self.languages)
        }

    @classmethod
    def from_allophone_data(
        cls,
        attribute_indexer: "PhoneticAttributeIndexer",
        languages: List[str],
    ):
        allophone_data = attribute_indexer.allophone_data
        if allophone_data is None:
            raise ValueError("No allophone data is available in the indexer")

        allophone_inventories = allophone_data.inventories
        shared_phone_indexer = allophone_data.shared_phone_indexer
        standardized_codes = [LanguageCode.from_str(code).alpha3 for code in languages]
        allophones = {}

        for language_id, language in enumerate(standardized_codes):
            allophone_inventory = (
                allophone_inventories.loc[allophone_inventories.ISO6393 == language, "Allophones"]
                .str.split(" ")
                .to_dict()
            )
            allophones[language_id] = {
                attribute_indexer.phoneme_index(phoneme): list(
                    map(int, shared_phone_indexer.phoneme_indices(allophones))
                )
                for phoneme, allophones in allophone_inventory.items()
            }

        return cls(allophones, languages, allophone_data.shared_phone_indexer.phonemes.tolist())


@dataclass
class LanguageInventories:
    inventories: Dict[int, List[str]]
    languages: List[str]

    def shared_inventory(self) -> List[str]:
        # Sorted for better reproducibility
        return sorted(set(itertools.chain.from_iterable(self.inventories.values())))

    def iso6393_inventories(self) -> Dict[str, List[str]]:
        return {
            LanguageCode.from_str(language).alpha3: self.inventories[language_id]
            for language_id, language in enumerate(self.languages)
        }

    def map_allophones(self, attribute_indexer: "PhonemeIndexer") -> LanguageAllophoneMappings:
        return LanguageAllophoneMappings(
            {
                language_id: {phoneme: [phoneme] for phoneme in map(int, attribute_indexer.phoneme_indices(inventory))}
                for language_id, inventory in self.inventories.items()
            },
            self.languages,
            attribute_indexer.phonemes.tolist(),
        )


@dataclass
class PhoneticIndexerState:
    phoneme_inventory: List[str]
    language_allophones: Optional[LanguageAllophoneMappings] = None
    table_file: Optional[str] = None


class PhonemeIndexer(metaclass=ABCMeta):
    _phoneme_data: DataFrame
    _feature_table: np.ndarray
    _feature_columns: Index
    _feature_names: List[str]
    _feature_categories: Dict[str, List[str]]

    @property
    def feature_table(self) -> np.ndarray:
        return self._feature_table

    @property
    def phoneme_data(self) -> DataFrame:
        return self._phoneme_data

    @property
    def phonemes(self) -> Index:
        return self._phoneme_data.index

    @property
    def feature_columns(self) -> Index:
        return self._feature_columns

    def phoneme_indices(self, phonemes: Sequence[str]) -> Tensor:
        phoneme_indices = self._phoneme_data.index.get_indexer(phonemes)
        if -1 in phoneme_indices:
            raise ValueError(f"Missing phonemes: {[phonemes[index] for index in np.where(phoneme_indices == -1)[0]]}")
        return LongTensor(phoneme_indices)

    def phoneme_indices_with_missing(self, phonemes: Sequence[str]) -> Tuple[Tensor, List[str]]:
        phoneme_indices = self._phoneme_data.index.get_indexer(phonemes)
        return LongTensor(phoneme_indices), [phonemes[index] for index in map(int, np.where(phoneme_indices == -1)[0])]

    def phoneme_index(self, phoneme: str) -> int:
        return typing.cast(int, self._phoneme_data.index.get_loc(phoneme))

    @overload
    def phoneme(self, index: int) -> str: ...

    @overload
    def phoneme(self, index: np.ndarray) -> Index: ...

    def phoneme(self, index: int | np.ndarray) -> str | Index:
        # Only compatible with np.ndarray and not torch tensors
        return typing.cast(str | Index, self._phoneme_data.index[index])

    def state(self) -> PhoneticIndexerState:
        return PhoneticIndexerState(self.phonemes.tolist())

    def segmenter(self) -> IpaSentenceSegmenter:
        return IpaSentenceSegmenter(self.phonemes.to_list())

    def __getitem__(self, index_or_name: int | Tensor | Tuple[int, Sequence[int] | int]) -> List[Tensor]:
        if isinstance(index_or_name, Tensor):
            index_or_name = index_or_name.numpy()

        feature_columns = self._feature_table[index_or_name].T
        # Special case for empty phoneme sequences
        if feature_columns.shape[1] == 0:
            return [torch.empty(0) for _ in range(len(feature_columns))]
        return [torch.from_numpy(np.concatenate(column)) for column in feature_columns]

    def get_named(
        self, index_or_name: List[str] | str | int | Tensor | np.ndarray, attribute_index_offset: int = 0
    ) -> Dict[str, Tensor]:
        if isinstance(index_or_name, list):
            indices = self.phoneme_indices(index_or_name).numpy()
        elif isinstance(index_or_name, str):
            indices = np.array([self.phoneme_index(index_or_name)])
        else:
            indices = index_or_name.numpy() if isinstance(index_or_name, Tensor) else index_or_name

        feature_matrix = self._feature_table[indices].T
        # Special case for empty phoneme sequences
        if feature_matrix.shape[1] == 0:
            return {name: torch.empty(0) for name in self._feature_columns}
        return {
            name: torch.from_numpy(np.concatenate(column)) + attribute_index_offset
            for name, column in zip(self._feature_columns, feature_matrix)
        }

    def feature_categories(self, feature: str) -> List[str]:
        return self._feature_categories[feature]

    def feature_category_index(self, name: str) -> int:
        return typing.cast(int, self._feature_columns.get_loc(name))

    def feature_values(self, name: str, feature_indices: Tensor) -> List[str]:
        categories = self._feature_categories[name]
        return [categories[index] for index in feature_indices]

    @property
    def feature_names(self) -> List[str]:
        return self._feature_names

    def __len__(self) -> int:
        # Phoneme count
        return self._feature_table.shape[0]


def _closest_phonemes(distance_matrix: Tensor, target_inventory: List[str]) -> Iterator[Tuple[str, int]]:
    """
    Gets the phonemes with the lowest distance for each row from a distance matrix

    :param distance_matrix: A distance matrix of size SxT where S is the number
        of source phonemes and T is the number of target phonemes. Note that T has
        to be equivalent toe the length of the `target_inventory`
    :param target_inventory: A list of phonemes in the target inventory where
        each element corresponds to a column position in the `distance_matrix`

    :return: An iterator over tuples of the closest target phonemes and their
        distance for each row in the distance matrix
    """
    return (
        (target_inventory[column_index], int(distance_matrix[row_index, column_index].item()))
        for row_index, column_index in enumerate(distance_matrix.argmin(1))
    )


PhonemeMapping = Dict[str, List[str]]


@dataclass
class DistanceMatrices:
    main: DataFrame
    splits: Dict[str, DataFrame]


class ArticulatoryAttributes(PhonemeIndexer):
    def __init__(
        self, phoneme_features: DataFrame, feature_categories: Dict[str, List[str]], reindex_phonemes: bool = True
    ):
        #  Removes phoneme from the feature set if present and keep_phonemes is False
        self._phoneme_data = phoneme_features

        if reindex_phonemes and "phoneme" in self._phoneme_data:
            # Reindex phonemes in subset order
            self._phoneme_data["phoneme"] = [np.array([index]) for index in range(len(self._phoneme_data))]
            feature_categories["phoneme"] = self._phoneme_data.index.tolist()

        self._feature_table = self._phoneme_data.values
        self._feature_columns = self._phoneme_data.columns
        # Filter feature categories with available columns, this also ensures
        # that the order should be the same as the columns in the matrix
        self._feature_categories = {name: feature_categories[name] for name in self._feature_columns}
        self._feature_names = self._feature_columns.tolist()

        # Extract the first feature of every contour for the hamming distance
        self._dense_feature_table = Tensor([contour[0] for contour in self._feature_table.flatten()]).reshape(
            self._feature_table.shape
        )

    @property
    def dense_feature_table(self) -> Tensor:
        return self._dense_feature_table

    def _phoneme_subset_fallback(self, subset_phonemes: DataFrame, phonemes: Sequence[str]) -> DataFrame:
        segmenter = self.segmenter().word_segmenter
        subset_indices, missing = self.phoneme_indices_with_missing(phonemes)
        # Replace missing indices with the indices of the first subsegments of missing phonemes
        missing_mask = subset_indices == -1
        if missing:
            subset_indices[missing_mask] = self.phoneme_indices(
                [segmenter.segment_checked(phoneme)[0] for phoneme in missing]
            )
        # Select subset
        subset_phonemes = subset_phonemes.iloc[subset_indices]
        # Replace phoneme representation with missing phonemes
        if missing:
            index = subset_phonemes.index.values.copy()
            index[missing_mask] = missing
            subset_phonemes.index = index

        return subset_phonemes

    def subset(
        self,
        phonemes: Sequence[str] | None = None,
        attribute_subset: Sequence[str] | None = None,
        reindex_phonemes: bool = True,
        missing_feature_fallback: bool = False,
    ):
        subset_phonemes = self._phoneme_data.copy()
        if phonemes is not None:
            if missing_feature_fallback:
                subset_phonemes = self._phoneme_subset_fallback(subset_phonemes, phonemes)
            else:
                subset_phonemes = subset_phonemes.iloc[self.phoneme_indices(phonemes)]
        if attribute_subset is not None:
            subset_phonemes = subset_phonemes[attribute_subset]
        # Copy subset to allow for further modification without affecting the superset
        return self.__class__(subset_phonemes, self._feature_categories.copy(), reindex_phonemes)

    def _hamming_distance(self, features: Tensor) -> Tensor:
        """
        Computes the hamming distances between a batch of feature vectors and
        the features of all available phonemes in the set

        :param feature: A matrix of phonetic feature vectors

        :return: A matrix of pairwise hamming distances between `features` and
            the phonetic feature vectors of all available phonemes
        """
        assert features.numel() > 0, "Cannot compute hamming distance without phonetic features"
        assert self._dense_feature_table.numel() > 0, "No features defined in the indexer"
        # cdist with p=0 corresponds to hamming distance
        return torch.cdist(features, self._dense_feature_table, 0)

    def _inventory_distance_matrix(
        self, source_inventory: List[str], target_inventory: List[str], add_base_count: bool = False
    ) -> Tensor:
        return torch.cdist(
            self._simplified_inventory_features(source_inventory, add_base_count),
            self._simplified_inventory_features(target_inventory, add_base_count),
            0,
        )

    @overload
    def map_inventories_closest(
        self,
        source_inventory: Sequence[str],
        target_inventory: Sequence[str],
        split_non_matching_complex: bool,
        return_distance_matrices: Literal[True],
        distance_threshold: int | None,
    ) -> Tuple[PhonemeMapping, DistanceMatrices]: ...

    @overload
    def map_inventories_closest(
        self,
        source_inventory: Sequence[str],
        target_inventory: Sequence[str],
        split_non_matching_complex: bool,
        return_distance_matrices: bool = False,
        distance_threshold: int | None = None,
    ) -> PhonemeMapping: ...

    def map_inventories_closest(
        self,
        source_inventory: Sequence[str],
        target_inventory: Sequence[str],
        split_non_matching_complex: bool = False,
        return_distance_matrices: bool = False,
        distance_threshold: int | None = None,
    ) -> Tuple[PhonemeMapping, DistanceMatrices] | PhonemeMapping:
        """
        Maps phonemes from the source inventory to the closest phonemes in the
        target inventory. Optionally splits complex segments and maps
        subsegments individually if the source phoneme is a complex segment and
        the closest target phoneme doesn't have the same number of subsegments.
        For instance, a diphthong that would be mapped to a single vowel will
        be split into its constituent vowels and each of its sub-vowels is
        mapped individually. For analysis purposes, distance matrices for the
        main inventory mapping and individual subsegment mappings can also be returned.

        :param source_inventory: A sequence of source phonemes to be mapped to the target
        :param target_inventory: The target phoneme inventory to map to
        :param split_non_matching_complex: Splits complex segments and maps
            subsegments individually if the source phoneme is a complex segment and
            the closest target phoneme doesn't have the same number of subsegments
        :param return_distance_matrices: Returns :py:class:`DistanceMatrices`
            in a tuple after phoneme mappings for the main inventory mapping and
            individual subsegment mappings if `split_non_matching_complex` is `True`
        :param distance_threshold: A phoneme mapping reverts to an identity mapping if the distance is higher or equal to the threshold.
            If complex segment splitting is active, the threshold applies to the subsegments after splitting
        :return: A :py:class:`PhonemeMapping` from the source to the target
            inventory or a tuple of a :py:class:`PhonemeMapping` and
            :py:class:`DistanceMatrices` if `return_distance_matrices` is `True`
        """
        # Map phonemes that match exactly first since different phonemes could
        # have the same features depending on the feature set and might map differently by chance otherwise
        matching = set(source_inventory).intersection(target_inventory)
        mapping = {phoneme: [phoneme] for phoneme in matching}
        # Only match source phonemes but use the full target inventory to allow for one-to-many mappings
        source_inventory = [phoneme for phoneme in source_inventory if phoneme not in matching]
        target_inventory = list(target_inventory)
        distance_matrix = self._inventory_distance_matrix(source_inventory, target_inventory, add_base_count=True)
        split_matrices = {}

        replacements = {}
        for phoneme, (target, distance) in zip(source_inventory, _closest_phonemes(distance_matrix, target_inventory)):
            phoneme_base = list(phoneme_segmentation.base_phonemes(phoneme))
            target_base = list(phoneme_segmentation.base_phonemes(target))
            # If splitting complex segments is disabled or the source and
            # target phonemes are both complex but with the same number of
            # subsegments (e.g. diphthong mapped to another diphthong)
            if not (split_non_matching_complex and len(phoneme_base) != len(target_base)):
                # Revert to an identity mapping if the distance is higher or equal to the threshold
                if distance_threshold is not None and distance >= distance_threshold:
                    target = phoneme
                # Direct source to target mapping
                replacements[phoneme] = [target]
                continue

            # Split complex segment if the target isn't a complex segment as well
            # Map subsegments individually instead
            subsegments = phoneme_segmentation.split_complex_segment(phoneme)
            split_matrix = self._inventory_distance_matrix(subsegments, target_inventory, add_base_count=True)
            # Store matrix for mapping subsegments if requested
            if return_distance_matrices:
                split_matrices[phoneme] = DataFrame(split_matrix, index=subsegments, columns=target_inventory)
            # Assign the replacement target phonemes with the lowest distance to each subsegment
            # Applies the threshold to each subsegment if provided
            replacements[phoneme] = [
                target if distance_threshold is None or distance < distance_threshold else subsegment
                for subsegment, (target, distance) in zip(
                    subsegments, _closest_phonemes(split_matrix, target_inventory)
                )
            ]

        # Merge distance based mappings with identity mappings for matching phonemes
        mapping.update(replacements)
        # Collect and warn about all target phonemes that weren't the closest phoneme to any source phoneme
        unmapped_from_target = set(target_inventory) - set(
            phoneme for phonemes in mapping.values() for phoneme in phonemes
        )
        if unmapped_from_target:
            logging.warn(f"{len(unmapped_from_target)} unmapped from target: {unmapped_from_target}")

        # Returns the general and split specific distance matrices if requested
        if return_distance_matrices:
            return (
                mapping,
                DistanceMatrices(
                    DataFrame(distance_matrix, index=source_inventory, columns=target_inventory), split_matrices
                ),
            )
        return mapping

    def closest_phone(self, features: Tensor) -> int:
        return int(self._hamming_distance(features.unsqueeze(0)).argmin())

    def closest_phone_for(self, phone: str, features: Tensor) -> int:
        # In cases where multiple phones have the same hamming distance and for performance
        if phone in self.phonemes:
            return self.phoneme_index(phone)
        return self.closest_phone(features.unsqueeze(0))

    def feature_vector(self, phone: str | int) -> np.ndarray:
        if isinstance(phone, str):
            phone = self.phoneme_index(phone)
        return self._feature_table[phone]

    def simplified_feature_vector(self, phone: str | int) -> Tensor:
        if isinstance(phone, str):
            phone = self.phoneme_index(phone)
        return self._dense_feature_table[phone]

    def _simplified_inventory_features(self, inventory: Sequence[str], add_base_count: bool = False) -> Tensor:
        features = self._dense_feature_table[self.phoneme_indices(inventory)]
        if not add_base_count:
            return features

        return torch.cat(
            (
                features,
                torch.tensor(
                    [utils.iterator_length(phoneme_segmentation.base_phonemes(segment)) for segment in inventory]
                ).unsqueeze(1),
            ),
            1,
        )

    def k_nearest_phones(self, phone_or_features: str | int | Tensor, k: int) -> Tensor:
        if not isinstance(phone_or_features, Tensor):
            phone_or_features = self.simplified_feature_vector(phone_or_features)

        # Stacks distance in the first row and indices in the second row
        return torch.stack(self._hamming_distance(phone_or_features.unsqueeze(0)).topk(k, largest=False))

    def missing_inventory_mappings(
        self, shared_inventory: Sequence[str], segment_missing: bool = False
    ) -> Dict[str, str]:
        """
        Generates a mapping between phonemes that are missing from the
        inventory due to differences in their unicode representation.

        Optionally, the mapping attempts to segment missing phonemes and chooses
        their first sub-segment only if the inventory contains all of its
        sub-segments, such as for unseen consonant clusters
        """
        segmenter = self.segmenter().word_segmenter
        mapping = {}
        for phoneme in shared_inventory:
            if phoneme not in self.phonemes:
                # Handles complications with unicode normalization
                combined_phoneme = unicodedata.normalize("NFC", phoneme)
                if combined_phoneme in self.phonemes:
                    mapping[phoneme] = combined_phoneme
                    continue

                if not segment_missing:
                    raise ValueError(f"No suitable mapping found for segment {phoneme!r}")

                # Segment phoneme and choose the first sub-segment if all
                # sub-segments are in the segmenter's phoneme inventory
                try:
                    mapping[phoneme] = segmenter.segment_checked(phoneme)[0]
                except MissingSegmentError as error:
                    # When no mapping with the other rules is possible
                    raise ValueError(f"No suitable mapping found for segment {phoneme!r}") from error

        return mapping


@dataclass
class AllophoneData:
    inventories: DataFrame
    shared_phone_indexer: ArticulatoryAttributes


def _extract_contours(column: Series):
    return column.str.split(",")


def _collect_vocabulary(column: Series, start_offset: int = 0) -> Dict[str, int]:
    return {value: index for index, value in enumerate(sorted(column.explode().unique()), start_offset)}


def _binarize_vocabulary(column: Series, vocabulary: Series) -> Series:
    current_vocabulary = vocabulary[column.name]

    def binarize_row(row: List[str]) -> np.ndarray:
        return np.array([current_vocabulary[element] for element in row], dtype=np.int64)

    return typing.cast(Series, column.apply(binarize_row))


FeatureTableInput = AnyPath | ReadCsvBuffer | str


def _binarize_contours(data: DataFrame, feature_start_column: str, vocabularies: Series | None = None) -> Series:
    attribute_contours = data.loc[:, feature_start_column:].apply(_extract_contours)
    if vocabularies is None:
        vocabularies = typing.cast(Series, attribute_contours.apply(_collect_vocabulary))

    data.loc[:, feature_start_column:] = attribute_contours.apply(_binarize_vocabulary, args=(vocabularies,))
    return vocabularies


LanguageInventoryTypes = LanguageInventories | LanguageAllophoneMappings | Sequence[str] | None


def generate_allophone_data(
    language_inventories: LanguageInventoryTypes,
    feature_table: DataFrame,
    attribute_subset: Sequence[str] | None = None,
    phoneme_subset: Sequence[str] | None = None,
) -> Tuple[DataFrame, Sequence[str]]:
    match language_inventories:
        case LanguageInventories():
            languages = language_inventories.languages
            inventories = language_inventories.iso6393_inventories()
        case LanguageAllophoneMappings():
            languages = language_inventories.languages
            if phoneme_subset is None:
                raise ValueError(
                    "allophone inventories can only be restored from LanguageAllophoneMappings if a correct phoneme_subset is provided"
                )
            inventories = language_inventories.iso6393_inventories(phoneme_subset)
        case None:
            languages = None
            inventories = None
        case language_codes:
            languages = language_codes
            inventories = None

    allophone_data = extract_allophone_inventories(
        feature_table.reset_index(),
        languages,
        attribute_subset,
        inventories,
        prefer_default_dialects=True,
        remove_zero_phoneme=True,
    ).set_index("phoneme")

    if phoneme_subset is None:
        phonemes: DataFrame = allophone_data[allophone_data["InventoryID"] != 0]
        phoneme_subset = phonemes.index.unique().tolist()

    return allophone_data, phoneme_subset


class PhoneticAttributeIndexer(PhonemeIndexer):
    def __init__(
        self,
        feature_set: FeatureSet,
        attribute_table_file: FeatureTableInput | None = None,
        attribute_subset: Sequence[str] | None = None,
        phoneme_subset: Sequence[str] | None = None,
        language_inventories: LanguageInventoryTypes = None,
        allophones_from_allophoible: bool = False,
    ):
        self._allophone_data = None

        if feature_set == FeatureSet.PHOIBLE:
            original_feature_table: DataFrame = read_allophoible(attribute_table_file, index_column="Phoneme")
            self._allophone_data, phoneme_subset = generate_allophone_data(
                language_inventories,
                original_feature_table,
                phoneme_subset=phoneme_subset,
            )

            feature_table = original_feature_table.copy(deep=True)
            feature_table.index.rename("phoneme", inplace=True)
            feature_start_column = "tone"
            phoneme_attributes = feature_table.loc[
                ~feature_table.index.duplicated(keep="first"),
                ["SegmentClass", *feature_table.columns[feature_table.columns.get_loc(feature_start_column) :]],
            ]
        elif feature_set == FeatureSet.PANPHON:
            if allophones_from_allophoible:
                raise NotImplementedError("Allophone handling is not implemented for Panphon features")
            original_feature_table = read_panphon(attribute_table_file, index_column="ipa")
            feature_start_column = "syl"
            # Removes erroneous duplicates vowels with delrel, the first strategy should only select those with delrel 0
            # See: https://github.com/dmort27/panphon/issues/26
            phoneme_attributes = original_feature_table.loc[
                ~original_feature_table.index.duplicated(keep="first"),
                original_feature_table.columns[original_feature_table.columns.get_loc(feature_start_column) :],
            ]
            # Add phoneme segments with their ties removed to handle G2P models that don't include ties
            rows_with_ties = phoneme_attributes[phoneme_attributes.index.str.contains(TIE)].copy()
            rows_with_ties.index = [segment.replace(TIE, "") for segment in rows_with_ties.index]
            phoneme_attributes = pd.concat([phoneme_attributes, rows_with_ties], verify_integrity=True)
            phoneme_attributes.index.rename("phoneme", inplace=True)
        else:
            raise ValueError(f"Unsupported feature set: {feature_set}")

        self._table_file = original_feature_table.to_csv()

        # Add phonemes as an extra column make all features categorical
        phoneme_attributes["phoneme"] = phoneme_attributes.index
        attribute_vocabularies = _binarize_contours(phoneme_attributes, feature_start_column)

        feature_categories = typing.cast(Dict[str, List[str]], attribute_vocabularies.apply(list).to_dict())
        # Full feature set used for general phoneme hamming distance
        self._full_attributes = ArticulatoryAttributes(
            phoneme_attributes.loc[:, feature_start_column:], feature_categories
        )
        # Features used for hamming distance which only includes the features
        # that are also used by the classifier - e.g. for handling inventory mapping
        # during zero-shot transfer
        self._subset_attributes = self._full_attributes.subset(phoneme_subset, attribute_subset)
        # Features with the reduced feature space but for all phonemes for generating evaluation labels with remapping
        if attribute_subset is None or "phoneme" in attribute_subset:
            full_subset = attribute_subset
        else:
            full_subset = [*attribute_subset, "phoneme"]
        self._full_phoneme_subset_attributes = self._full_attributes.subset(attribute_subset=full_subset)

        self._phoneme_data = self._subset_attributes.phoneme_data.copy()

        # Create indexing information for the subset used during training and validation
        self._feature_categories = self._subset_attributes._feature_categories
        self._feature_table = self._subset_attributes.feature_table
        self._feature_columns = self._subset_attributes.feature_columns
        self._feature_names = self._subset_attributes.feature_names

        self._feature_counts = LongTensor([len(self._feature_categories[name]) for name in self._feature_columns])
        self._total_size = int(self._feature_counts.sum().item())

        # Start column for the feature subset
        start_column = "tone" if feature_set == FeatureSet.PHOIBLE else "syl"
        feature_start_column = typing.cast(
            str,
            self._full_attributes._feature_columns[
                typing.cast(int, self._full_attributes._feature_columns.get_loc(start_column)) + 1
            ],
        )
        varying_feature_categories = self._full_attributes._feature_categories.copy()
        if feature_set == FeatureSet.PHOIBLE:
            del varying_feature_categories[start_column]

        # Keep track of all phonetic features required for embedding
        # composition by removing the "phoneme" column
        features_only = varying_feature_categories.copy()
        features_only.pop("phoneme", None)
        self._composition_features = list(features_only)

        if self._allophone_data is not None:
            _binarize_contours(self._allophone_data, feature_start_column, attribute_vocabularies)

            self._allophone_data = AllophoneData(
                self._allophone_data,
                ArticulatoryAttributes(
                    self._allophone_data.loc[~self._allophone_data.index.duplicated(keep="first"), feature_start_column:],
                    varying_feature_categories,
                ),
            )

        # Create mappings from inventories if they are given
        match language_inventories:
            case LanguageAllophoneMappings():
                self._language_allophones = language_inventories
            case LanguageInventories():
                if allophones_from_allophoible:
                    self._language_allophones = LanguageAllophoneMappings.from_allophone_data(
                        self,
                        language_inventories.languages,
                    )
                else:
                    self._language_allophones = language_inventories.map_allophones(self)
            case _:
                self._language_allophones = None

        if self._language_allophones is not None:
            self._feature_categories["phone"] = self._language_allophones.shared_phones

    def state(self) -> PhoneticIndexerState:
        return PhoneticIndexerState(self.phonemes.tolist(), self._language_allophones, self._table_file)

    @classmethod
    def from_state(
        cls,
        feature_set: FeatureSet,
        state: PhoneticIndexerState,
        feature_subset: List[str] | None = None,
    ):
        return cls(
            feature_set,
            state.table_file,
            feature_subset,
            state.phoneme_inventory,
            # Always initialize with allophone data if phoible features are used
            allophones_from_allophoible=feature_set == FeatureSet.PHOIBLE,
        )

    @classmethod
    def from_config(
        cls,
        config: Config,
        attribute_table_file: FeatureTableInput | None = None,
        language_inventories: LanguageInventories | None = None,
        state_dict: PhoneticIndexerState | None = None,
    ):
        # Makes use of the guarantee that dict preserves insertion order to keep attributes ordered while set doesn't preserve order
        existing_entries = {}
        for entry in config.nn.projection.classes:
            existing_entries[entry.name] = None
            existing_entries.update((attribute, None) for attribute in entry.dependencies)

        # Remove output dependency tag since it is not a valid feature
        existing_entries.pop(ProjectionEntryConfig.OUTPUT_DEPENDENCY, None)
        for attribute in list(existing_entries):
            if ProjectionEntryConfig.OUTPUT_PATTERN.match(attribute):
                del existing_entries[attribute]

        match state_dict:
            case PhoneticIndexerState(
                phoneme_inventory, table_file=table_file, language_allophones=LanguageAllophoneMappings()
            ):
                language_allophone_mappings = state_dict.language_allophones
                phoneme_subset = phoneme_inventory
                attribute_table_file = table_file
            case _ if language_inventories is not None:
                language_allophone_mappings = language_inventories
                phoneme_subset = sorted(language_inventories.shared_inventory())
            case _:
                language_allophone_mappings = phoneme_subset = None

        return cls(
            config.nn.projection.feature_set,
            attribute_table_file,
            list(existing_entries.keys()),
            phoneme_subset,
            language_allophone_mappings,
            config.nn.projection.phoneme_layer == PhonemeLayerType.ALLOPHONES,
        )

    @property
    def composition_features(self) -> List[str]:
        return self._composition_features

    @property
    def language_allophones(self) -> LanguageAllophoneMappings | None:
        return self._language_allophones

    @property
    def attributes(self) -> ArticulatoryAttributes:
        return self._subset_attributes

    @property
    def full_attributes(self) -> ArticulatoryAttributes:
        return self._full_attributes

    @property
    def full_subset_attributes(self) -> ArticulatoryAttributes:
        return self._full_phoneme_subset_attributes

    def composition_feature_matrix(self, inventory: list[str]) -> Tensor:
        """
        Constructs the feature matrix used by the :py:class:`allophant.network.acoustic_model.EmbeddingCompositionLayer`
        for recognizing phones from a given inventory

        :param inventory: An inventory of phones supported by the Allophoible database

        :return: A matrix of feature vectors for each phone in the inventory.
            In the case of contour features, the first feature value in the contour is used
        """
        return self._full_attributes.subset(inventory, self._composition_features).dense_feature_table.long()

    def allophone_inventory(self, language_code: str) -> DataFrame:
        if self._allophone_data is None:
            raise ValueError(
                'Allophone inventories can only be accessed if features were extracted from Allophoible'
            )

        return self._allophone_data.inventories[
            self._allophone_data.inventories.ISO6393 == language_codes.standardize_to_iso6393(language_code)
        ]

    def phoneme_inventory(self, languages: Sequence[str] | str) -> list[str]:
        """
        Constructs a (shared) phoneme inventory for the given language or
        multiple languages. If a sequence of language codes is given, the union
        of phoneme inventories is returned.

        :param languages: Either a single language code or a sequence of
            language codes.

        :return: A list of phonemes from the given language or languages. If a
            sequence of language codes is given, the list contains the union of
            phoneme inventories.
        """
        if self._allophone_data is None:
            raise ValueError(
                'Allophone inventories can only be accessed if features were extracted from Allophoible'
            )

        if isinstance(languages, str):
            selection = self._allophone_data.inventories.ISO6393 == language_codes.standardize_to_iso6393(languages)
        else:
            selection = self._allophone_data.inventories.ISO6393.isin(
                {language_codes.standardize_to_iso6393(language_code) for language_code in languages}
            )

        return self._allophone_data.inventories[selection].index.unique().to_list()

    @overload
    def map_language_inventory(
        self,
        inventories: Iterable[List[str]],
        language: str,
        return_distance_matrices: Literal[True],
        distance_threshold: int | None,
    ) -> List[Tuple[PhonemeMapping, DistanceMatrices]]: ...

    @overload
    def map_language_inventory(
        self,
        inventories: Iterable[List[str]],
        language: str,
        return_distance_matrices: bool = False,
        distance_threshold: int | None = None,
    ) -> List[PhonemeMapping]: ...

    def map_language_inventory(
        self,
        inventories: Iterable[List[str]],
        language: str,
        return_distance_matrices: bool = False,
        distance_threshold: int | None = None,
    ) -> List[Tuple[PhonemeMapping, DistanceMatrices]] | List[PhonemeMapping]:
        phoneme_inventory = self.allophone_inventory(language).index.tolist()
        # Remap all phoneme inventories with the full allophoible
        # If the target inventory is smaller than the source inventory,
        # remaining phonemes are mapped to the closest one in the target inventory
        return [
            self.full_attributes.map_inventories_closest(
                inventory,
                phoneme_inventory,
                split_non_matching_complex=True,
                return_distance_matrices=return_distance_matrices,
                distance_threshold=distance_threshold,
            )
            for inventory in inventories
        ]

    @property
    def allophone_data(self) -> AllophoneData | None:
        return self._allophone_data

    def size(self, column: int | str | None = None) -> int:
        if column is None:
            return self._total_size
        if isinstance(column, str):
            column = self.feature_category_index(column)
        return int(self._feature_counts[column].item())

    def map_to_subset(self, inventory: Sequence[str]) -> Dict[str, str]:
        current_segments = self._subset_attributes
        inventory_segments = self._full_attributes.subset(inventory)

        return {
            phoneme: current_segments.phoneme(
                current_segments.closest_phone_for(phoneme, inventory_segments.simplified_feature_vector(phoneme))
            )
            for phoneme in inventory
        }

    def _phoneme_fallback(self, segmenter: IpaSegmenter, phoneme: str) -> str:
        if phoneme in self.phonemes:
            return phoneme

        return segmenter.segment_checked(phoneme)[0]

    def map_target_inventory(
        self,
        inventory: Sequence[str],
        map_uncovered_target_phonemes: bool = True,
        missing_feature_fallback: bool = False,
    ) -> Dict[str, str]:
        """
        Maps a given source inventory to the closest phones in the indexer
        based on the "tr2tgt" scheme from Xu et al. (2022).

        :param inventory: The source inventory to map to the closest phones in the indexer
        :param map_uncovered_target_phonemes: Enables mapping the closest phone
            in the source inventory to any uncovered target phone after the initial
            mapping phase as in the original "tr2tgt" scheme
        :param missing_feature_fallback: Falls back to features of the first
            subsegment of complex phones for which no features are available to
            complete the mapping

        :return: A mapping from the given source inventory to the indexer's target inventory

        .. references::
            Xu, Q., Baevski, A., Auli, M. (2022) Simple and Effective Zero-shot
            Cross-lingual Phoneme Recognition. Proc. Interspeech 2022,
            2113-2117, doi: 10.21437/Interspeech.2022-60
        """
        current_segments = self._subset_attributes
        inventory_segments = self._full_attributes.subset(inventory, missing_feature_fallback=missing_feature_fallback)
        remaining_phonemes = set(inventory)

        source_mapping = {}
        for phoneme in self.phonemes:
            target_phoneme = inventory_segments.phoneme(
                inventory_segments.closest_phone_for(phoneme, current_segments.simplified_feature_vector(phoneme))
            )
            source_mapping[phoneme] = target_phoneme
            remaining_phonemes.discard(target_phoneme)

        if map_uncovered_target_phonemes:
            for target_phoneme in remaining_phonemes:
                closest_in_source = current_segments.phoneme(
                    current_segments.closest_phone_for(
                        target_phoneme, inventory_segments.simplified_feature_vector(target_phoneme)
                    )
                )
                source_mapping[closest_in_source] = target_phoneme

        return source_mapping


def read_panphon(
    file: FeatureTableInput | None = None, index_column: str | None = None, use_categorical: bool = False
) -> DataFrame:
    column_dtype = CategoricalDtype(ordered=True) if use_categorical else str
    if file is not None:
        if isinstance(file, str):
            file = StringIO(file)
        panphon_features = pd.read_csv(file, dtype=column_dtype, index_col=index_column)
    else:
        with (resources.files(panphon) / "data/ipa_all.csv").open("r", encoding="utf-8") as default_file:
            panphon_features = pd.read_csv(default_file, dtype=column_dtype, index_col=index_column)

    return panphon_features


def read_allophoible(file: FeatureTableInput | None = None, index_column: str | None = None) -> DataFrame:
    if file is not None:
        if isinstance(file, str):
            file = StringIO(file)
        allophoible = pd.read_csv(file, dtype=str, index_col=index_column)
    else:
        with ALLOPHOIBLE_PATH.open("r", encoding="utf-8") as default_file:
            allophoible = pd.read_csv(default_file, dtype=str, index_col=index_column)

    return allophoible.astype(
        {"InventoryID": int, "SegmentClass": CategoricalDtype(["consonant", "vowel", "tone", "null"])}
    )


class LanguageMappingWarning(UserWarning):
    """Warns about languages being remapped to a closely related variant"""


class SingletonFeatureWarning(UserWarning):
    """Warns about features not varying"""


warnings.simplefilter("always", LanguageMappingWarning)
warnings.simplefilter("always", SingletonFeatureWarning)


_SOURCE_AND_LANGUAGE = ["Source", "ISO6393", "SpecificDialect"]


def _select_largest_inventories(
    non_marginal_allophones: DataFrame, preferred_dialects: Dict[str, str] | None = None
) -> DataFrame:
    data = non_marginal_allophones[_SOURCE_AND_LANGUAGE]
    if preferred_dialects is not None:
        # Selects preferred dialects and remove other dialects with the same language code
        data = pd.concat(
            [
                *(
                    data[(data.ISO6393 == language) & (data.SpecificDialect == dialect)]
                    for language, dialect in preferred_dialects.items()
                ),
                data[~data.ISO6393.isin(preferred_dialects)],
            ]
        )

    # Select the largest phoneme inventory per language
    return (
        data.groupby(_SOURCE_AND_LANGUAGE, dropna=False)
        .size()
        .sort_values(ascending=False)  # type: ignore
        .reset_index()
        .drop_duplicates("ISO6393")
    )


def _filter_inventory(
    phoible: DataFrame, remapped_inventories: Dict[str, List[str]]
) -> Callable[[DataFrame], DataFrame]:
    def inventory_filter(inventory: DataFrame) -> DataFrame:
        expected_inventory = set(remapped_inventories[inventory.name])
        inventory_subset = inventory[inventory["Phoneme"].isin(expected_inventory)]
        remaining_phonemes = expected_inventory - set(inventory_subset["Phoneme"])
        if not remaining_phonemes:
            return inventory_subset

        remaining_inventory = phoible[phoible["Phoneme"].isin(remaining_phonemes)].drop_duplicates("Phoneme").copy()
        # Assign only the phonemes as allophones
        remaining_inventory["Allophones"] = remaining_inventory["Phoneme"]
        remaining_inventory.loc[:, "InventoryID":"SpecificDialect"] = inventory_subset.loc[
            [inventory_subset.index[0]] * len(remaining_inventory), "InventoryID":"SpecificDialect"
        ].values
        remaining_inventory["Marginal"] = None
        assert len(remaining_inventory) == len(remaining_phonemes), "Inventory mismatch detected"
        return pd.concat((inventory_subset, remaining_inventory))

    return inventory_filter


def extract_allophone_inventories(
    phoible: DataFrame,
    language_codes: Sequence[str] | None = None,
    attribute_subset: Sequence[str] | None = None,
    remapped_inventories: Dict[str, List[str]] | None = None,
    prefer_default_dialects: bool = False,
    remove_zero_phoneme: bool = False,
) -> DataFrame:
    non_marginal_allophones: DataFrame = phoible[~phoible["Allophones"].isna() & (phoible["Marginal"] != "TRUE")]
    if language_codes is not None:
        language_codes_iso6393 = {LanguageCode.from_str(code).alpha3 for code in language_codes}
        filtered: DataFrame = non_marginal_allophones[non_marginal_allophones.ISO6393.isin(language_codes_iso6393)]
    else:
        language_codes_iso6393 = None
        filtered = non_marginal_allophones.copy()

    if prefer_default_dialects:
        with DEFAULT_DIALECTS_PATH.open("r", encoding="utf-8") as file:
            default_dialects = json.load(file)
    else:
        default_dialects = None

    languages = _select_largest_inventories(filtered, default_dialects)

    # Try resolving language codes with no inventories from PHOIBLE to their
    # macro language codes to find the correct inventory
    if language_codes_iso6393 is not None and len(languages) != len(language_codes_iso6393):
        phoible_language_codes = non_marginal_allophones.ISO6393.unique()
        missing_languages = {
            LanguageCode.from_str(language, True, True).alpha3_t: language
            for language in set(language_codes_iso6393) - set(languages.ISO6393)
        }
        missing_mappings = {}

        for language in phoible_language_codes:
            macro = LanguageCode.from_str(language, True, True).alpha3_t
            if macro in missing_languages:
                missing_mappings[missing_languages.pop(macro)] = language
            # Prefer languages that are already macro languages over variants even if a variant was already mapped
            elif language == macro and macro in missing_mappings:
                missing_mappings[missing_mappings[macro]] = language

        if missing_languages:
            raise ValueError(
                f"Some of the requested languages don't contain allophone data: {sorted(missing_languages.values())}"
            )

        warnings.warn(
            f"Remapped some languages to a variant within the same macro language: {missing_mappings}",
            LanguageMappingWarning,
        )
        languages = pd.concat(
            (
                languages,
                _select_largest_inventories(
                    non_marginal_allophones[non_marginal_allophones.ISO6393.isin(missing_mappings.values())],
                    default_dialects,
                ),
            )
        )
    else:
        missing_mappings = {}

    # Filter rows for all source and language pairs that were filtered in the previous step
    filtered = phoible[
        phoible.set_index(_SOURCE_AND_LANGUAGE).index.isin(languages.set_index(_SOURCE_AND_LANGUAGE).index)
    ].copy()

    # Replaces remapped languages
    filtered.ISO6393 = filtered.ISO6393.replace({macro: language for language, macro in missing_mappings.items()})

    # Filter again using remapped inventories if they are provided
    if remapped_inventories is not None:
        filtered = filtered.groupby("ISO6393").apply(_filter_inventory(phoible, remapped_inventories))

    if remove_zero_phoneme:
        # Filters out all null phonemes including surrounding whitespace in a list
        filtered["Allophones"].replace(r"( ?∅|∅ ?)", "", regex=True, inplace=True)

    unique_allophones = filtered["Allophones"].str.split(" ").explode().unique()
    unique_phonemes = set(filtered["Phoneme"].unique())
    missing_phonemes = set(unique_allophones) - unique_phonemes
    additional_phones = phoible[phoible["Phoneme"].isin(missing_phonemes)].drop_duplicates("Phoneme")

    missing_features = missing_phonemes - set(additional_phones["Phoneme"])
    if missing_features:
        raise ValueError(
            "Missing pre-computed feature definitions for", len(missing_features), "allophones:", missing_features
        )

    # Set inventory for additional phones to the unused 0 index and remove language specific information that is left over as artifacts
    additional_phones["InventoryID"] = 0
    additional_phones.loc[:, "Glottocode":"SpecificDialect"] = pd.NA
    additional_phones.loc[:, ["Source", "Allophones"]] = pd.NA

    phoible_subset = pd.concat((additional_phones, filtered))
    # Remove columns that aren't required for the feature table
    phoible_subset = phoible_subset.drop(["Marginal"], axis=1)

    # Set phonemes as the index and rename for compatibility with other lower case phonetic feature names
    phoible_subset.rename(columns={"Phoneme": "phoneme"}, inplace=True)

    last_non_feature_column = typing.cast(int, phoible_subset.columns.get_loc("Source"))
    # Filter attribute subset
    if attribute_subset is not None:
        initial_columns = phoible_subset.columns[: last_non_feature_column + 1].tolist()
        if "phoneme" in initial_columns:
            try:
                initial_columns.remove("phoneme")
            except ValueError:
                pass

        initial_columns.extend(attribute_subset)
        phoible_subset = phoible_subset.loc[:, initial_columns]

    singleton_columns = phoible_subset.iloc[:, last_non_feature_column + 1 :].nunique() <= 1
    if singleton_columns.any():
        warnings.warn(
            f"Only one feature variant found in {phoible_subset.iloc[:, last_non_feature_column + 1:].columns[singleton_columns].tolist()}",
            SingletonFeatureWarning,
        )

    return phoible_subset


def main(args: Sequence[str] | None = None) -> None:
    if args is None:
        args = sys.argv[1:]

    parser = ArgumentParser(description="Extracts the most suitable PHOIBLE inventories for every language or a subset")
    parser.add_argument(
        "language_codes",
        nargs="?",
        type=lambda codes: codes.split(","),
        help="ISO639 language codes for which to extract inventories",
    )
    parser.add_argument(
        "-o",
        "--out",
        type=FileType("w", encoding="utf-8"),
        default=sys.stdout,
        help="Output paths for the processed PHOIBLE CSV",
    )
    parser.add_argument(
        "-p",
        "--feature-path",
        type=FileType("r", encoding="utf-8"),
        help="Path to a custom PHOIBLE version to process instead of the included version",
    )
    parser.add_argument(
        "-r", "--remove-zero", action="store_true", help="Removes the zero phoneme from allophone positions"
    )
    parser.add_argument(
        "-d",
        "--prefer-allophant-dialects",
        action="store_true",
        help=(
            "Selects the same dialect used during Allophant pre-training over using the largest inventory "
            "for some languages with inventories for multiple dialects"
        ),
    )

    arguments = parser.parse_args(args)

    with ALLOPHOIBLE_PATH.open("r", encoding="utf-8") if not arguments.feature_path else arguments.feature_path as file:
        allophoible = read_allophoible(file)

    with arguments.out as file:
        extract_allophone_inventories(
            allophoible,
            arguments.language_codes,
            None,
            None,
            arguments.prefer_allophant_dialects,
            arguments.remove_zero,
        ).to_csv(file, index=False)


if __name__ == "__main__":
    main(sys.argv[1:])



from typing import Tuple
import typing

import torch
from torch import Tensor
from torch.nn import Module
from torch.nn import functional


def get_padding(kernel_size: int, stride: int = 1, stft_type: bool = False) -> Tuple[int, int]:
    if stft_type:
        padding = kernel_size // 2
        if stride == 1:
            return (padding, padding - 1)
        else:
            return (padding, padding)
    if stride > 1:
        # Adds enough padding for a filter with a higher stride to reach the edge
        return (kernel_size // 2, kernel_size - 1)
    padding = kernel_size // 2
    return (padding, padding)


class VariableLengthReflectPad(Module):
    def __init__(self, padding: Tuple[int, int]):
        super().__init__()
        self._padding = padding
        self._left_padding, self._right_padding = padding
        pad_base_indices = torch.arange(0, self._right_padding).view(1, 1, -1)
        right_pad_start_indices = pad_base_indices + self._left_padding
        right_pad_end_indices = pad_base_indices + 2
        left_pad_indices = torch.arange(self._left_padding, 0, -1)
        self.register_buffer("_right_pad_start_indices", right_pad_start_indices)
        self.register_buffer("_right_pad_end_indices", right_pad_end_indices)
        self.register_buffer("_left_pad_indices", left_pad_indices)

    @property
    def padding(self) -> Tuple[int, int]:
        return self._padding

    def forward(self, features: Tensor, lengths: Tensor) -> Tensor:
        padded = functional.pad(features, self._padding)
        feature_size = features.size(1)
        padded[..., : self._left_padding] = features.gather(
            -1, typing.cast(Tensor, self._left_pad_indices).repeat(1, feature_size, 1)
        )
        lengths = lengths.view(-1, 1, 1)
        padded.scatter_(
            -1,
            (lengths + self._right_pad_start_indices).repeat(1, feature_size, 1),
            features.gather(-1, (lengths - self._right_pad_end_indices).repeat(1, feature_size, 1)),
        )
        return padded



from collections.abc import Callable
import typing
from typing import Tuple

import torch
from torch import Tensor
from torch import nn
from torch.nn import functional
from torchaudio.transforms import MFCC, MelSpectrogram

from allophant.utils import Squeeze, Unsqueeze, LogCompression
from allophant.config import Config, FeatureType, Window
from allophant.network import padding


class PreEmphasis(nn.Module):
    def __init__(self, coefficient: float = 0.97, trainable: bool = False):
        super().__init__()
        self.coefficient = coefficient
        # Uses symmetric 1 padding though only initial padding is relevant
        self._pre_emphasis_convolution = nn.Conv1d(1, 1, kernel_size=2, bias=False)
        with torch.no_grad():
            # Initializes to [-0.97, 1] which is equivalent to the regular pre-emphasis formula
            self._pre_emphasis_convolution.weight.copy_(torch.tensor([-coefficient, 1]))
        if not trainable:
            self._pre_emphasis_convolution.requires_grad_(False)
        self._padding = padding.get_padding(2, stft_type=True)

    def forward(self, batch: Tensor) -> Tensor:
        if batch.ndim == 3:
            return self._pre_emphasis_convolution(functional.pad(batch, self._padding, "reflect"))
        # Reflect padding to match lengths of in/out
        return self._pre_emphasis_convolution(functional.pad(batch.view(1, 1, -1), self._padding, "reflect")).ravel()


def next_power_of_2(x):
    return 1 if x == 0 else 2 ** (x - 1).bit_length()


def frame_count(num_samples: int, stride_samples: int) -> int:
    # Chosen to be equivalent to the torch.sfft implementation
    return 1 + (num_samples // stride_samples)


def _mono_channel(audio: Tensor) -> Tensor:
    # Assumes channel first
    return audio[0]


class FeatureFunction:
    def __init__(
        self,
        name: str,
        function: Callable[[Tensor], Tensor],
        sample_rate: int,
        dimensions: Tuple[int | None, int] | Tuple[int, int | None] | None = None,
        window: Window | None = None,
    ):
        self._name = name
        self._function = function
        self._stride_samples = window.frame_stride_samples(sample_rate) if window is not None else None
        # Transpose if lengths are specified before features
        has_dimensions = dimensions is not None
        self._transpose_required = False
        if has_dimensions:
            # Transpose dimensions if necessary
            if dimensions[1] is None:
                self._transpose_required = True
                dimensions = (None, dimensions[0])
            self._max_dimensions = typing.cast(Tuple[None, int], dimensions)
            self._feature_size = self._max_dimensions[1]
            self._dimensions = (1, self._feature_size)
        else:
            self._max_dimensions = (None,)
            self._feature_size = 1
            self._dimensions = (1,)

    @property
    def name(self) -> str:
        return self._name

    @property
    def feature_size(self) -> int:
        return self._feature_size

    @property
    def dimensions(self) -> Tuple[int, int] | Tuple[int]:
        return self._dimensions

    def frame_count(self, num_samples: int) -> int:
        if not self._stride_samples:
            return num_samples
        return frame_count(num_samples, self._stride_samples)

    def __call__(self, waveform: Tensor) -> Tensor:
        processed = self._function(waveform)
        if self._transpose_required:
            return processed.transpose(-2, -1)
        return processed

    @classmethod
    def from_config(cls, config: Config, sample_rate: int):
        feature_type = config.preprocessing.feature_type
        num_filters = config.preprocessing.num_filters
        n_fft = next_power_of_2(int((1024 / 16_000) * sample_rate))

        if feature_type == FeatureType.MFCC:
            window = config.preprocessing.window
            layers = [
                PreEmphasis(),
                MFCC(
                    sample_rate,
                    n_mfcc=num_filters,
                    log_mels=False,
                    melkwargs={
                        "n_fft": n_fft,
                        "win_length": window.frame_duration_samples(sample_rate),
                        "hop_length": window.frame_stride_samples(sample_rate),
                    },
                ),
                Unsqueeze(0),
                # CMVN
                nn.InstanceNorm1d(num_filters),
                Squeeze(0),
            ]
            return cls(
                feature_type.value,
                nn.Sequential(*layers),
                sample_rate,
                (num_filters, None),
                window,
            )
        elif feature_type == FeatureType.FILTERBANKS:
            window = config.preprocessing.window
            layers = [
                PreEmphasis(),
                MelSpectrogram(
                    sample_rate,
                    n_fft=n_fft,
                    win_length=window.frame_duration_samples(sample_rate),
                    hop_length=window.frame_stride_samples(sample_rate),
                    n_mels=num_filters,
                ),
                LogCompression(),
                Unsqueeze(0),
                # CMVN
                nn.InstanceNorm1d(num_filters),
                Squeeze(0),
            ]
            return cls(
                feature_type.value,
                nn.Sequential(*layers),
                sample_rate,
                (num_filters, None),
                window,
            )
        elif feature_type == FeatureType.RAW:
            # Just raw audio without a filter
            return cls("raw", _mono_channel, sample_rate)

        raise ValueError(f"Unsupported feature type: {feature_type}")



from torch import Tensor
from torch import nn
from torch.nn import functional
from abc import ABCMeta, abstractmethod


class LossWrapper(nn.Module, metaclass=ABCMeta):
    _loss: nn.Module

    @property
    def wrapped_loss(self):
        return self._loss

    @abstractmethod
    def forward(self, logits: Tensor, labels: Tensor, predicted_lengths: Tensor, label_lengths: Tensor) -> Tensor:
        pass


class CTCWrapper(LossWrapper):
    def __init__(self):
        super().__init__()
        # Enables zero losses if the label sequence is longer than the input
        # sequence to be robust against outliers
        self._loss = nn.CTCLoss(reduction="sum", zero_infinity=True)

    def forward(self, logits: Tensor, labels: Tensor, predicted_lengths: Tensor, label_lengths: Tensor) -> Tensor:
        return self._loss(functional.log_softmax(logits, -1), labels, predicted_lengths, label_lengths)


class SequenceCrossEntropyWrapper(LossWrapper):
    def __init__(self, label_smoothing: float = 0):
        super().__init__()
        # Enables zero losses if the label sequence is longer than the input
        # sequence to be robust against outliers
        self._loss = nn.CrossEntropyLoss(reduction="sum", label_smoothing=label_smoothing)

    def forward(
        self,
        logits: Tensor,
        labels: Tensor,
        _predicted_lengths: Tensor | None = None,
        _label_lengths: Tensor | None = None,
    ) -> Tensor:
        return self._loss(logits, labels)



from collections.abc import Callable, Iterable, Sequence
import random
import typing
from typing import Any, Iterator, List, TypeVar

from torch import Generator, Tensor, LongTensor
import torch
from torch.utils.data import BatchSampler, DataLoader, RandomSampler, Sampler, SequentialSampler
from torch.nn.utils import rnn
import numpy as np
from allophant import MAIN_LOGGER, utils

from allophant.config import BatchingMode
from allophant.dataset_processing import Batch, BatchType, LabeledBatch, PhonemeDataset, RawLabeledBatch
from allophant.datasets.speech_corpus import MultilingualCorpus


class _SequentialSamplerWrapper:
    """Wrapper class for the _LanguageBinSampler to get a consistent signature with `RandomSampler`"""

    def __init__(self, indices: range, generator: Generator | None = None) -> None:
        self._sampler = SequentialSampler(indices)

    def __iter__(self) -> Iterator[int]:
        yield from self._sampler


class _LanguageBinSampler:
    def __init__(self, language_indices: range, shuffle: bool = False, generator: Generator | None = None) -> None:
        self._sampler_type = RandomSampler if shuffle else _SequentialSamplerWrapper
        self._generator = generator
        self._language_indices = language_indices
        self._sampler = iter(self._sampler_type(language_indices, generator=self._generator))
        self._start_offset = language_indices.start

    def _next_sample(self) -> int:
        return next(self._sampler) + self._start_offset

    def sample(self) -> int:
        try:
            return self._next_sample()
        except StopIteration:
            # Recreate the sample without replacement from the start again
            self._sampler = iter(self._sampler_type(self._language_indices, generator=self._generator))
            return self._next_sample()


class LanguageOversamplingSampler(Sampler[int]):
    """
    Sampler using language oversampling based on XLS-R pre-training (Babu et al., 2021)

    .. references::
        Babu, Arun, Changhan Wang, Andros Tjandra, Kushal Lakhotia,
        Qiantong Xu, Naman Goyal, Kritika Singh et al. "XLS-R:
        Self-supervised cross-lingual speech representation learning at
        scale." arXiv preprint arXiv:2111.09296 (2021).
    """

    def __init__(
        self,
        corpus: MultilingualCorpus,
        oversampling_factor: float = 0.5,
        shuffle: bool = False,
        generator: Generator | None = None,
    ) -> None:
        self._language_bins = []
        utterance_counts = []
        for language in corpus.languages:
            indices = corpus.monolingual_index_range(language)
            self._language_bins.append(_LanguageBinSampler(indices, shuffle, generator))
            utterance_counts.append(len(indices))

        utterance_counts = torch.tensor(utterance_counts, dtype=torch.float32)
        utterance_weights = (utterance_counts / utterance_counts.sum()) ** oversampling_factor
        highest_resource_index = utterance_weights.argmax()
        highest_resource_factor = utterance_counts[highest_resource_index] / utterance_weights[highest_resource_index]
        # Gets the number of samples by ensuring that all utterances of the highest resource language are sampled at least once
        # and utterances from all other languages according to the distribution
        expected_samples = (highest_resource_factor * utterance_weights).round().long()
        bin_indices = torch.repeat_interleave(torch.arange(len(expected_samples)), expected_samples)

        self._samples_per_epoch = int(expected_samples.sum().item())
        self._bin_indices = bin_indices
        self._language_sampler = RandomSampler(bin_indices, generator=generator)

    def __iter__(self) -> Iterator[int]:
        for language_index in self._language_sampler:
            yield self._language_bins[self._bin_indices[language_index]].sample()

    def __len__(self) -> int:
        return self._samples_per_epoch


class MaxFrameBatchSampler(BatchSampler):
    """
    :py:class:`~torch.utils.data.BatchSampler` implementation for batching
    tensors together until the product of their sequence and batch dimension
    would grow larger than a maximum number of frames. This ensures consistent memory usage
    where batches use as much memory as possible below a maximum
    """

    def __init__(self, sampler: Sampler[int] | Iterable[int], batch_size: int, frame_lengths: Tensor) -> None:
        """
        Creates a batch sampler with the given batch size in maximum number of
        frames and a tensor of lengths for every sequence in the dataset.

        :param sampler: A :py:class:`~torch.utils.data.Sampler` for generating (random) data indices
        :param batch_size: The maximum number of frames that the product of the batch and sequence dimension cannot exceed
        :param frame_lengths: A 1d tensor of lengths for every sample in the dataset.
            The tensor should be at least as long as the largest index that can be generated from the `sampler`
        """
        self._sampler = sampler
        self._batch_size = batch_size
        self._frame_lengths = frame_lengths

    def __iter__(self) -> Iterator[List[int]]:
        batch_indices = []
        max_length = 0

        for index in self._sampler:
            length = self._frame_lengths[index]
            if length > max_length:
                max_length = length

            # Predicted batch size dependent on the dense size if the current word was included
            batch_size = len(batch_indices)
            new_batch_size = (batch_size + 1) * max_length
            # Return the current batch if the batch size would be too large with the current word
            if new_batch_size > self._batch_size:
                yield batch_indices
                # New batch with the current word which wasn't included in the last one
                max_length = length
                batch_indices = [index]
            else:
                batch_indices.append(index)

        # Handle final batch if necessary
        if batch_indices:
            yield batch_indices


class SkipBatchSampler(BatchSampler):
    def __init__(self, sampler: BatchSampler, skip_count: int) -> None:
        self._sampler = sampler
        self._skip_count = skip_count

    def __iter__(self) -> Iterator[List[int]]:
        samples = iter(self._sampler)
        if self._skip_count == 0:
            return samples

        skipped = 0
        for indices, _ in zip(samples, range(self._skip_count)):
            skipped += len(indices)

        if skipped:
            MAIN_LOGGER.info(f"Skipped {skipped}")

        return samples


def _build_batch(batch_type: BatchType) -> Callable[[Sequence[Batch]], Batch]:
    """
    Builds a dense batch from a sequence of single entry batches

    :param entries: A sequence of batches where each batch only contains a single entry

    :return: A dense batch with all entries merged together in (padded) tensors
    """

    def _create_batch(entries: Sequence[T]) -> Batch:
        lengths = LongTensor([entry.lengths for entry in entries])
        language_ids = LongTensor([entry.language_ids for entry in entries])
        audio_features = rnn.pad_sequence([entry.audio_features for entry in entries], True)
        # Only transpose if there is a feature dimension
        if audio_features.ndim > 2:
            audio_features = audio_features.transpose(1, 2)
        match batch_type:
            case BatchType.UNLABELED:
                return Batch(audio_features, lengths, language_ids)
            case BatchType.RAW:
                # Redefine entries with a narrower type - required since Sequence isn't invariant
                labeled_entries = typing.cast(Sequence[RawLabeledBatch], entries)
                return RawLabeledBatch(
                    audio_features,
                    lengths,
                    language_ids,
                    [list(labels) for labels in zip(*(entry.raw_labels[0] for entry in labeled_entries))],
                    [entry.utterance_ids[0] for entry in labeled_entries],
                )

        # Redefine entries with a narrower type - required since Sequence isn't invariant
        labeled_entries = typing.cast(Sequence[LabeledBatch], entries)
        label_lengths = []
        attribute_indices = []

        if entries:
            num_engines = len(labeled_entries[0].attribute_indices)
            for engine in range(num_engines):
                label_lengths.append(torch.stack([entry.label_lengths[engine] for entry in labeled_entries], 1))
                attribute_indices.append(
                    {
                        indices[0][0]: rnn.pad_sequence([tensor for _, tensor in indices], True)
                        for indices in zip(*(entry.attribute_indices[engine].items() for entry in labeled_entries))
                    }
                )

        return LabeledBatch(
            audio_features,
            lengths,
            language_ids,
            attribute_indices,
            label_lengths,
            labeled_entries[0].label_length_indices if entries else {},
        )

    return _create_batch


def _seed_worker(_worker_id: int):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


T = TypeVar("T", bound=Batch)


class Batcher:
    """
    Manages generation of batches for prediction and validation using utterance or frame
    count based batch sizes to ensure consistent and efficient memory usage
    """

    def __init__(
        self,
        batch_size: int,
        batching_mode: BatchingMode,
        language_oversampling_factor: float | None = None,
        data_workers: int | None = 0,
    ):
        """
        Initializes a `Batcher`

        :param batch_size: The maximum number of utterance per batch in "utterances" mode
            or the maximum number of speech frames per batch including padding in "frames" mode
        :param batching_mode: A :py:class:`~allophant.config.BatchingMode` for choosing between utterance based or frames based batch sizes
        :param language_oversampling_factor: Factor to use for oversampling from lower resource languages. Oversampling is based on XLS-R pre-training (Babu et al., 2021)
        :param data_workers: Number of worker threads used in batch generation - if `None` is
            specified the number of threads is inferred from the number available in the CPU

        .. references::
            Babu, Arun, Changhan Wang, Andros Tjandra, Kushal Lakhotia,
            Qiantong Xu, Naman Goyal, Kritika Singh et al. "XLS-R:
            Self-supervised cross-lingual speech representation learning at
            scale." arXiv preprint arXiv:2111.09296 (2021).
        """
        self._batch_size = batch_size
        self._batching_mode = batching_mode
        self._language_oversampling_factor = language_oversampling_factor
        self._data_workers = utils.get_worker_count(data_workers)

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def batches(
        self,
        data: PhonemeDataset[Any, Any, T],
        data_lengths: Tensor | None = None,
        shuffle: bool = False,
        seed: int | None = None,
        skip_batches: int = 0,
    ) -> Iterator[T]:
        """
        Generates batches given the size and mode defined in the :py:class:`Batcher`. During batching, the data is optionally shuffled.
        If the batching mode is set to use the maximum number of frames as the criterion, lengths for each utterance in the dataset need to be provided.

        :param data: The dataset for which to generate batches
        :param data_lengths: A :py:class:`~torch.Tensor` which is only required by the frame count based batching strategy
        :param shuffle: Whether to shuffle the data while batching
        :param seed: Sets a seed for the random shuffling of data when `shuffle` is enabled
        :param skip_batches: Skips the first `n` batches

        :return: A generator over labeled or unlabeled batches depending on the properties of the :py:class:`Batcher`
        """
        corpus = data.corpus
        random_generator = Generator()
        if seed is not None:
            random_generator.manual_seed(seed)

        if self._batching_mode == BatchingMode.UTTERANCES:
            if self._language_oversampling_factor is None:
                sampler = SequentialSampler(corpus)
            else:
                sampler = LanguageOversamplingSampler(
                    corpus,
                    self._language_oversampling_factor,
                    shuffle,
                    random_generator,
                )
            yield from DataLoader(
                data,
                shuffle=shuffle,
                batch_sampler=SkipBatchSampler(BatchSampler(sampler, self._batch_size, drop_last=False), skip_batches),
                collate_fn=_build_batch(data.batch_type),
                num_workers=self._data_workers,
                persistent_workers=True,
                worker_init_fn=None if seed is None else _seed_worker,
                generator=random_generator,
            )
            return

        if data_lengths is None:
            raise ValueError("Frame Lengths for each utterance are required for using max frame batching")

        if self._language_oversampling_factor is None:
            sampler = RandomSampler(corpus, generator=random_generator) if shuffle else SequentialSampler(corpus)
        else:
            sampler = LanguageOversamplingSampler(
                corpus,
                self._language_oversampling_factor,
                shuffle,
                random_generator,
            )

        yield from DataLoader(
            data,
            batch_sampler=SkipBatchSampler(
                MaxFrameBatchSampler(
                    sampler,
                    self._batch_size,
                    data_lengths,
                ),
                skip_batches,
            ),
            collate_fn=_build_batch(data.batch_type),
            num_workers=self._data_workers,
            persistent_workers=True,
            worker_init_fn=None if seed is None else _seed_worker,
            generator=random_generator,
        )



from collections.abc import Sequence
from dataclasses import dataclass
import os
from os import path
from typing import Dict, Iterable, Iterator, List

from tqdm import tqdm

from allophant.datasets import MultilingualCorpus
from allophant.datasets.phonemes import LanguageCodeAny
from allophant import language_codes
from allophant.datasets.speech_corpus import LanguageData, LanguageInfo, PhoneticallySegmentedUtterance
from allophant.utils import AnyPath


@dataclass
class Transcription:
    """
    A transcription from the :py:class:`UCLAPhoneticCorpus`

    :param utterance_id: The ID of the utterance
    :param raw: The narrow phone annotations of the utterance
    :param phonemes: The segmented and normalized transcription from the utterance
    """

    utterance_id: str
    raw: str
    phonemes: List[str]


@dataclass
class UCLALanguageData(LanguageData[LanguageInfo, Transcription]):
    """
    Metadata and transcribed utterances for a monolingual subset of the :py:class:`UCLAPhoneticCorpus`

    :param info: Language-specific metadata
    :param transcribed_samples: A list of transcribed utterances for a language
    :param id_map: A map from utterance IDs to generated utterance indices to allow access by ID
    """

    id_map: Dict[str, int]


_INVENTORY_FILE = "inventory"
_RAW_FILE = "raw"
_TEXT_FILE = "text"


def _load_languages(
    directory: AnyPath, languages: Sequence[LanguageCodeAny] | None = None, progress_bar: bool = False
) -> Iterator[UCLALanguageData]:
    """
    Loads transcriptions and inventories from the full
    :py:class:`UCLAPhoneticCorpus` or a subset with fewer languages

    :param directory: Base directory of the corpus
    :param languages: A language subset or `None` to load the full corpus
    :param progress_bar: If true, loading progress is displayed as a progress bar

    :return: A generator over monolingual subsets of the :py:class:`UCLAPhoneticCorpus`
    """
    if languages is None:
        language_generator = os.listdir(directory)
        total = len(language_generator)
    else:
        language_generator = (language_codes.to_language_code(code).alpha3 for code in languages)
        total = len(languages)

    for language in (
        tqdm(language_generator, total=total, unit=" languages", desc="Loading transcriptions")
        if progress_bar
        else language_generator
    ):
        language_directory = path.join(directory, language)

        with open(path.join(language_directory, _INVENTORY_FILE), "r", encoding="utf-8") as file:
            # Sort inventories for consistency with other processing functions
            info = LanguageInfo(language, sorted(phoneme for phoneme, _ in map(str.split, file)), [])

        with (
            open(path.join(language_directory, _RAW_FILE), "r", encoding="utf-8") as raw_file,
            open(path.join(language_directory, _TEXT_FILE), "r", encoding="utf-8") as text_file,
        ):
            transcriptions = []
            id_map = {}
            for i, (raw_line, text_line) in enumerate(zip(raw_file, text_file)):
                raw_id, raw = raw_line.split()
                text_columns = text_line.split()
                assert raw_id == text_columns[0], "Mismatch between raw and text file IDs"
                transcriptions.append(Transcription(raw_id, raw, text_columns[1:]))
                id_map[raw_id] = i

        yield UCLALanguageData(info, transcriptions, id_map)


class UCLAPhoneticCorpus(MultilingualCorpus):
    """
    An interface to the utterances transcriptions and inventories contained in the UCLA Phonetic Corpus introduced by Li et al. (2021)

    .. references::
        Li, X., Mortensen, D. R., Metze, F., & Black, A. W. (2021, June).
        Multilingual phonetic dataset for low resource speech recognition.
        In ICASSP 2021-2021 IEEE International Conference on Acoustics, Speech and
        Signal Processing (ICASSP) (pp. 6958-6962). IEEE.
    """

    UTTERANCE_TYPE = PhoneticallySegmentedUtterance

    _SAMPLE_RATES = [44100, 48000]
    _AUDIO_DIRECTORY = "audio"

    def __init__(self, base_directory: AnyPath, languages: Iterable[LanguageData], resample: int | None = None) -> None:
        """
        Creates a new UCLA Phonetic Corpus from the given data and base directory

        :param base_directory: Base directory of the corpus
        :param language_data: An iterable yielding `LanguageData` instances which
            contain metadata and a list of utterances for monolingual subsets from the corpus.
            If multiple `LanguageData` instances are provided, a `ValueError` will be raised.
        :param resample: If given, all utterances will be resampled to the given sample rate when they are loaded
        """
        super().__init__(base_directory, languages, self._AUDIO_DIRECTORY, "wav", resample=resample)

    @classmethod
    def load(
        cls,
        directory: AnyPath,
        resample: int | None = None,
        languages: Sequence[LanguageCodeAny] | None = None,
        progress_bar: bool = False,
    ):
        """
        Loads (a subset of) the UCLA Phonetic Corpus from the given directory

        :param directory: Base directory of the corpus
        :param resample: If given, all utterances will be resampled to the given sample rate when they are loaded
        :param languages: A language subset or `None` to load the full corpus
        :param progress_bar: If true, loading progress is displayed as a progress bar
        """
        return cls(directory, _load_languages(directory, languages, progress_bar), resample)

    def __str__(self) -> str:
        return (
            f"UCLA Phonetic Corpus containing {len(self._languages)} languages with {self._num_utterances} utterances"
        )



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import Dict, Tuple, TypeVar, Type, List, Optional
import math
import typing

import torch
from torch import LongTensor, autograd, nn
from torch.nn import functional
from torch import Tensor
from torch.nn.parameter import Parameter
from transformers import Wav2Vec2Config
from transformers.models.wav2vec2.modeling_wav2vec2 import Wav2Vec2Model
from transformers.models.wav2vec2.feature_extraction_wav2vec2 import Wav2Vec2FeatureExtractor

from allophant.attribute_graph import AttributeGraph, AttributeNode
from allophant.config import (
    Architecture,
    EmbeddingCompositionConfig,
    PhonemeLayerType,
    ProjectionConfig,
    ProjectionEntryConfig,
    TransformerAcousticModelConfig,
    UnfreezeScheduleConfig,
    Wav2Vec2PretrainedConfig,
)
from allophant.network import frontend
from allophant.network.frontend import Frontend, SequentialFrontend, frontend_from_config
from allophant.batching import Batch
from allophant import MAIN_LOGGER, utils
from allophant.phonetic_features import ArticulatoryAttributes, LanguageAllophoneMappings, PhoneticAttributeIndexer


class SinusoidalPositionEmbeddings(nn.Module):
    """
    Sinusoidal positional embeddings from Vaswani et al. (2017). Adapted from
    the implementation by Rush (2018)

    .. references::
        Vaswani, Ashish, Noam M. Shazeer, Niki Parmar, Jakob Uszkoreit, Llion
        Jones, Aidan N. Gomez, Lukasz Kaiser and Illia Polosukhin. “Attention
        is All you Need.” Neural Information Processing Systems (2017).

        Rush, Alexander. “The Annotated Transformer.” In Proceedings of
        Workshop for NLP Open Source Software (NLP-OSS), 52–60. Melbourne,
        Australia: Association for Computational Linguistics, 2018.
        https://doi.org/10.18653/v1/W18-2509.
    """

    _LOG_10000 = math.log(10000)

    _bases: Tensor

    def __init__(self, input_size: int):
        super().__init__()
        component = torch.exp(torch.arange(0, input_size, 2, dtype=torch.float) * -(self._LOG_10000 / input_size))
        self.register_buffer("_bases", torch.stack([component] * 2, 1).view(-1), persistent=False)
        self.embedding_size = input_size

    def get_positions(self, max_positions: int) -> Tensor:
        device = self._bases.device
        position_embeddings = torch.arange(max_positions, dtype=torch.float, device=device).unsqueeze(1) * self._bases
        position_embeddings[:, 0::2] = torch.sin(position_embeddings[:, 0::2])
        position_embeddings[:, 1::2] = torch.cos(position_embeddings[:, 1::2])

        return position_embeddings.view(max_positions, -1)

    def forward(self, batch: Tensor) -> Tensor:
        return batch + self.get_positions(batch.size(0)).unsqueeze(1)


_PAD_VALUE = torch.finfo(torch.float32).min


@torch.jit.script  # pyright: ignore
def _multiply_allophone_matrix(
    phone_logits: Tensor, matrix: Tensor, mask: Tensor, mask_value: float = _PAD_VALUE
) -> Tensor:
    return (
        (phone_logits * matrix.unsqueeze(0))
        .masked_fill_(
            mask.unsqueeze(0),
            mask_value,
        )
        .max(1)
        .values
    )


class AllophoneMapping(nn.Module):
    """
    Allophone layer derived from the Allosaurus architecture (Li et al., 2020).

    .. references::
        Li, Xinjian, Siddharth Dalmia, Juncheng Li, Matthew Lee, Patrick
        Littell, Jiali Yao, Antonios Anastasopoulos, et al. “Universal Phone
        Recognition with a Multilingual Allophone System.” In ICASSP 2020 -
        2020 IEEE International Conference on Acoustics, Speech and Signal
        Processing (ICASSP), 8249–53. Barcelona, Spain: IEEE, 2020.
        https://doi.org/10.1109/ICASSP40776.2020.9054362.
    """

    _allophone_mask: Tensor

    def __init__(
        self,
        shared_phone_count: int,
        phoneme_count: int,
        blank_offset: int,
        language_allophones: LanguageAllophoneMappings,
    ) -> None:
        super().__init__()
        allophones = language_allophones.allophones
        languages = language_allophones.languages
        num_languages = len(languages)
        # Maps language codes to dense indices in the matrix
        self._index_map = {}

        allophone_matrix = torch.zeros(num_languages, shared_phone_count, phoneme_count)
        for dense_index, (language_index, allophone_mapping) in enumerate(allophones.items()):
            language_allophone_matrix = allophone_matrix[dense_index]
            # Constructs identity mappings for BLANK in the diagonal
            language_allophone_matrix[range(blank_offset), range(blank_offset)] = 1

            self._index_map[languages[language_index]] = dense_index
            for phoneme, allophones in allophone_mapping.items():
                # Add allophone mappings with the blank offset
                language_allophone_matrix[torch.tensor(allophones) + blank_offset, phoneme + blank_offset] = 1

        # Shared allophone_matrix
        self._allophone_matrices = Parameter(allophone_matrix)

        # Initialization for the l2 penalty according to "Universal Phone Recognition With a Multilingual Allophone Systems" by Li et al.
        self.register_buffer("_initialization", allophone_matrix.clone(), persistent=False)
        # Inverse mask for language specific phoneme inventories
        self.register_buffer("_allophone_mask", ~allophone_matrix.bool(), persistent=False)

    @property
    def index_map(self) -> Dict[str, int]:
        return self._index_map

    def map_allophones(self, phone_logits: Tensor, language_ids: Tensor) -> Tensor:
        # Batch size x Shared Phoneme Inventory Size
        batch_matrices = torch.empty(
            *phone_logits.shape[:2], self._allophone_matrices.shape[2], device=phone_logits.device
        )
        for index, language_id in enumerate(map(int, language_ids)):
            logits = phone_logits[:, index].unsqueeze(-1)
            # Max pooling after multiplying with the allophone matrix
            # Replace masked positions with negative infinity
            # since this results in zero probabilities after softmax for phone and
            # phoneme combinations that don't occur in the allophone mappings
            batch_matrices[:, index] = _multiply_allophone_matrix(
                logits,
                self._allophone_matrices[language_id],
                self._allophone_mask[language_id],
            )

        return batch_matrices

    def forward(self, phone_logits: Tensor, language_ids: Tensor, predict: bool = False) -> Dict[str, Tensor]:
        # Note that while predictions on the training corpus could still be valid with the allophone layer enabled,
        # language IDs are different for other corpora and therefore not supported
        if predict:
            # Also assigns phone logits to the phoneme layer in case another layer uses phonemes as a dependency
            return {ProjectionEntryConfig.PHONE: phone_logits, ProjectionEntryConfig.PHONEME_LAYER: phone_logits}
        return {ProjectionEntryConfig.PHONEME_LAYER: self.map_allophones(phone_logits, language_ids)}

    def l2_penalty(self) -> Tensor:
        """
        Computes the l2 penalty for the allophone layer to be added to the loss

        :return: The l2 penalty for the allophone layer given the current
            weights or 0 if no allophone layer is present in the architecture
        """
        # Calculates the matrix L2 (Frobenius) norm for each allophone matrix and then sums the norms over languages
        return torch.norm_except_dim(self._allophone_matrices - self._initialization, dim=0).sum()


class EmbeddingCompositionLayer(nn.Module):
    """
    Embedding composition layer derived from the compositional phone embedding
    layer by Li et al, (2021)

    .. references::
        Li, Xinjian, Juncheng Li, Florian Metze and Alan W. Black.
        “Hierarchical Phone Recognition with Compositional Phonetics.”
        Interspeech (2021).
    """

    def __init__(self, embedding_size: int, attribute_indexer: ArticulatoryAttributes) -> None:
        super().__init__()

        dense_feature_table = attribute_indexer.dense_feature_table.long()
        # Add Single blank embedding
        num_categories = torch.cat((LongTensor([0]), dense_feature_table.max(0).values)) + 1
        unused_categories = torch.cat(
            (torch.tensor([False]), torch.cat([row.bincount() for row in dense_feature_table.T]) == 0)
        )

        unused_count = unused_categories.sum().item()
        if unused_count:
            MAIN_LOGGER.info(f"{unused_count} unused feature embeddings")

        # Offsets and one additional entry for the special blank feature
        category_offsets = num_categories.cumsum(0)[:-1].unsqueeze(0)
        dense_feature_table += category_offsets
        self._attribute_embeddings = nn.EmbeddingBag(int(num_categories.sum()), embedding_size, mode="sum")

        # Set unused attribute embedding weights to 0
        with torch.no_grad():
            self._attribute_embeddings.weight[unused_categories] = 0

        self.register_buffer("_dense_feature_table", dense_feature_table, persistent=False)
        self.register_buffer("_category_offsets", category_offsets, persistent=False)
        # Scale factor for the dot product with feature embeddings as in Li et al. (2021)
        self.register_buffer("_scale_factor", torch.tensor(math.sqrt(embedding_size)), persistent=False)

    def forward(self, inputs: Tensor, target_feature_indices: Tensor | None = None) -> Tensor:
        if target_feature_indices is None:
            target_feature_indices = typing.cast(Tensor, self._dense_feature_table)
        else:
            target_feature_indices = target_feature_indices + self._category_offsets

        composed_embeddings = torch.cat(
            (
                # Blank embedding (Using the index batch sequence equivalent to [[0]])
                self._attribute_embeddings(torch.zeros(1, 1, dtype=target_feature_indices.dtype, device=inputs.device)),
                # Phonemes
                self._attribute_embeddings(target_feature_indices),
            )
        ).T

        return (inputs @ composed_embeddings) / self._scale_factor


class ProjectingMultiheadAttention(nn.Module):
    def __init__(
        self,
        input_dimensions: int,
        hidden_dimensions: int,
        num_heads: int,
        add_positional_embeddings: bool = False,
        dropout_rate: float = 0,
    ) -> None:
        super().__init__()
        self.input_projection = nn.Linear(input_dimensions, hidden_dimensions)
        self.positional_embeddings = (
            SinusoidalPositionEmbeddings(hidden_dimensions) if add_positional_embeddings else None
        )
        self.layer_norm = nn.LayerNorm(hidden_dimensions)
        self.attention = nn.MultiheadAttention(hidden_dimensions, num_heads)
        self.dropout = nn.Dropout(dropout_rate)

    def forward(self, inputs: Tensor, lengths: Tensor) -> Tensor:
        projection_output = self.layer_norm(self.input_projection(inputs))
        if self.positional_embeddings is not None:
            projection_output = self.positional_embeddings(projection_output)

        return self.dropout(
            self.attention(
                projection_output,
                projection_output,
                projection_output,
                utils.mask_sequence(lengths, inverse=True),
                need_weights=False,
            )[0]
        )


class HierarchicalClassifier(nn.Module):
    def __init__(
        self,
        time_distributed_layer: ProjectingMultiheadAttention | nn.Linear,
        composition_layer: EmbeddingCompositionLayer | None = None,
        allophone_layer: AllophoneMapping | None = None,
    ) -> None:
        super().__init__()
        self._lengths_required = not isinstance(time_distributed_layer, nn.Linear)
        self._time_distributed_layer = time_distributed_layer
        self._composition_layer = composition_layer
        self._allophone_layer = allophone_layer

    def forward(
        self,
        inputs: Tensor,
        lengths: Tensor,
        language_ids: Tensor,
        target_feature_indices: Tensor | None = None,
        predict: bool = False,
    ) -> Tensor:
        if self._lengths_required:
            inputs = self._time_distributed_layer(
                inputs,
                lengths,
            )
        else:
            inputs = self._time_distributed_layer(inputs)

        if self._composition_layer is not None:
            inputs = self._composition_layer(inputs, target_feature_indices)

        if self._allophone_layer is None:
            return inputs

        return self._allophone_layer(inputs, language_ids, predict)


def _process_classifier_dependencies(
    attribute_graph: AttributeGraph,
    node: AttributeNode,
    output_features: int,
    blank_offset: int,
    dependency_blanks: bool = True,
) -> Tuple[int, List[AttributeNode]]:
    layer_input_neurons = 0
    dependencies = []
    for target in node.dependencies:
        attribute_node = attribute_graph.get(target)
        if attribute_node is None:
            # Output nodes aren't in the graph and are turned into a node for convenience
            attribute_node = AttributeNode(target, output_features)
        elif dependency_blanks:
            # Add the blank offset only to phonetic attribute nodes if necessary
            attribute_node = attribute_node.with_offset(blank_offset)

        layer_input_neurons += attribute_node.size
        dependencies.append(attribute_node)

    return layer_input_neurons, dependencies


class HierarchicalProjection(nn.Module):
    _OUTPUT_PATTERN = ProjectionEntryConfig.OUTPUT_PATTERN

    def __init__(
        self,
        output_features: int,
        attribute_graph: AttributeGraph,
        blank_offset: int,
        dependency_blanks: bool = True,
        language_allophones: LanguageAllophoneMappings | None = None,
        attribute_indexer: PhoneticAttributeIndexer | None = None,
        acoustic_model_dropout_rate: float = 0,
        embedding_composition_config: EmbeddingCompositionConfig | None = None,
    ):
        super().__init__()
        self._acoustic_model_dropout = (
            nn.Dropout(acoustic_model_dropout_rate) if acoustic_model_dropout_rate > 0 else None
        )
        self._uses_allophone_mapping = False

        dependency_names = set(attribute_graph.names())
        if len(dependency_names) < len(attribute_graph):
            raise ValueError("Dependencies contain duplicate keys")
        if any(self._OUTPUT_PATTERN.match(name) for name in dependency_names):
            raise ValueError(f"{ProjectionEntryConfig.OUTPUT_DEPENDENCY!r} is a reserved keyword")

        self._blank_offset = blank_offset
        self._dependency_blanks = dependency_blanks

        self._layers = nn.ModuleDict()
        self._ordered_nodes = []

        required_output_layers = set()

        # Generate layers by iterating over the attribute graph in reverse topological order
        for node in attribute_graph.sort():
            # Retrieve dependency nodes
            layer_input_neurons, dependencies = _process_classifier_dependencies(
                attribute_graph,
                node,
                output_features,
                blank_offset,
                dependency_blanks,
            )

            if not dependencies:
                raise ValueError("Each class projection requires a dependency")

            # Keep track of topologically ordered graph nodes with their dependencies for the forward pass
            self._ordered_nodes.append((node.name, dependencies))
            # Keep track of required acoustic model output layers
            required_output_layers.update(
                dependency.name for dependency in dependencies if self._OUTPUT_PATTERN.match(dependency.name)
            )

            # Use an allophone layer
            is_phoneme_layer = node.name == ProjectionEntryConfig.PHONEME_LAYER
            if language_allophones is not None and is_phoneme_layer:
                self._uses_allophone_mapping = True
                # Use the size of the phone inventory as an input including CTC
                # offset as passthrough to allow for decoding even though shared phones aren't explicitly supervised
                output_size = len(language_allophones.shared_phones) + blank_offset
            else:
                # Get the output size plus space for CTC blanks
                output_size = node.size + blank_offset

            if is_phoneme_layer and embedding_composition_config is not None:
                # Project to the embedding size when using embedding composition
                projection_output_size = embedding_composition_config.embedding_size
            else:
                # Project to the number of classes otherwise
                projection_output_size = output_size

            if node.time_layer_config is not None:
                time_distributed_layer = ProjectingMultiheadAttention(
                    layer_input_neurons,
                    projection_output_size,
                    node.time_layer_config.num_heads,
                    node.time_layer_config.positional_embeddings,
                    acoustic_model_dropout_rate,
                )
            else:
                time_distributed_layer = nn.Linear(layer_input_neurons, projection_output_size)

            if is_phoneme_layer and embedding_composition_config is not None:
                if attribute_indexer is None:
                    raise ValueError(
                        "Model configuration using attribute embedding composition"
                        " requires an attribute indexer but got `None`"
                    )

                if not self._uses_allophone_mapping:
                    # Use the phoneme subset from the attribute indexer with all features for constructing the embeddings
                    training_attributes = attribute_indexer.full_attributes.subset(
                        attribute_indexer.phonemes.tolist(),
                        attribute_indexer.composition_features.copy(),
                    )
                else:
                    if attribute_indexer.allophone_data is None:
                        raise ValueError(
                            "Model configuration using attribute embedding composition and an allophone layer"
                            " requires allophone data in the attribute indexer with but got `None`"
                        )

                    training_attributes = attribute_indexer.allophone_data.shared_phone_indexer

                if output_size != len(training_attributes) + 1:
                    raise ValueError(
                        f"Length of attributes with blanks ({len(training_attributes) + 1}) need to match"
                        f" the number of phones in the allophone mapping ({output_size})"
                    )

                embedding_size = embedding_composition_config.embedding_size
                composition_layer = EmbeddingCompositionLayer(embedding_size, training_attributes)
            else:
                composition_layer = None

            if is_phoneme_layer and self._uses_allophone_mapping:
                allophone_layer = AllophoneMapping(
                    output_size,
                    # The node contains the number of language specific phonemes
                    node.size + blank_offset,
                    blank_offset,
                    typing.cast(LanguageAllophoneMappings, language_allophones),
                )
            else:
                allophone_layer = None

            self._layers[node.name] = HierarchicalClassifier(time_distributed_layer, composition_layer, allophone_layer)

        if not required_output_layers:
            raise ValueError(
                "At least one of the input layers requires {ProjectionEntryConfig.OUTPUT_DEPENDENCY!r} as a dependency"
            )

        # Sorted for consistency
        self._output_dependencies = sorted(required_output_layers)

    def forward(
        self,
        inputs: List[Tensor],
        input_lengths: Tensor,
        language_ids: Tensor,
        target_feature_indices: Tensor | None = None,
        predict: bool = False,
    ) -> Dict[str, Tensor]:
        outputs = {
            f"{ProjectionEntryConfig.OUTPUT_DEPENDENCY}_{index}": input_layer
            for index, input_layer in enumerate(inputs)
        }
        outputs[ProjectionEntryConfig.OUTPUT_DEPENDENCY] = inputs[-1]

        # Apply dropout to the acoustic model outputs if necessary
        if self._acoustic_model_dropout is not None:
            for dependency in self._output_dependencies:
                outputs[dependency] = self._acoustic_model_dropout(outputs[dependency])

        projection_outputs = {}
        # Iterate over ordered nodes and layer - note that ModuleDict maintains insertion order
        for (name, dependencies), layer in zip(self._ordered_nodes, self._layers.values()):
            # Fast path for a single output dependency to avoid concatenation
            if len(dependencies) == 1 and self._OUTPUT_PATTERN.match(dependencies[0].name):
                dependency_outputs = outputs[dependencies[0].name]
            else:
                dependency_outputs = torch.cat(
                    [
                        (
                            torch.softmax(
                                (
                                    outputs[dependency.name]
                                    if self._dependency_blanks
                                    else outputs[dependency.name][..., self._blank_offset :]
                                ),
                                -1,
                            )
                            if not self._OUTPUT_PATTERN.match(dependency.name)
                            else outputs[dependency.name]
                        )
                        for dependency in dependencies
                    ],
                    -1,
                )

            output = layer(dependency_outputs, input_lengths, language_ids, target_feature_indices, predict)
            if isinstance(output, dict):
                projection_outputs.update(output)
                outputs.update(output)
            else:
                projection_outputs[name] = output
                outputs[name] = output

        return projection_outputs

    def l2_penalty(self) -> Tensor | None:
        """
        Computes the l2 penalty for the allophone layer to be added to the loss

        :return: The l2 penalty for the allophone layer given the current
            weights or 0 if no allophone layer is present in the architecture
        """
        try:
            allophone_mapping = self._layers[ProjectionEntryConfig.PHONEME_LAYER]
            if isinstance(allophone_mapping, AllophoneMapping):
                return allophone_mapping.l2_penalty()
        finally:
            # No penalty if there's no allophone layer in the architecture
            return None

    def map_allophones(self, phone_logits: Tensor, language_ids: Tensor) -> Tensor:
        allophone_mapping = self._layers[ProjectionEntryConfig.PHONEME_LAYER]
        if isinstance(allophone_mapping, AllophoneMapping):
            return allophone_mapping.map_allophones(phone_logits, language_ids)

        raise ValueError("Can't map phones to allophones with a model without an allophone layer")

    @property
    def classifier_layers(self) -> nn.ModuleDict:
        return self._layers


# NOTE: Taken from https://pytorch.org/docs/stable/_modules/torch/nn/modules/transformer.html#TransformerEncoderLayer
def _get_activation_fn(activation):
    if activation == "relu":
        return functional.relu
    elif activation == "gelu":
        return functional.gelu

    raise RuntimeError("activation should be relu/gelu, not {}".format(activation))


# NOTE: Modified from https://pytorch.org/docs/stable/_modules/torch/nn/modules/transformer.html#TransformerEncoderLayer
class PreLMTransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        nhead: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: str = "relu",
        elementwise_affine: bool = False,
    ):
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model, elementwise_affine=elementwise_affine)
        self.norm2 = nn.LayerNorm(d_model, elementwise_affine=elementwise_affine)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)

    def __setstate__(self, state):
        if "activation" not in state:
            state["activation"] = functional.relu
        super().__setstate__(state)

    def forward(
        self, src: Tensor, src_mask: Optional[Tensor] = None, src_key_padding_mask: Optional[Tensor] = None
    ) -> Tensor:
        r"""Pass the input through the encoder layer.

        Args:
            src: the sequence to the encoder layer (required).
            src_mask: the mask for the src sequence (optional).
            src_key_padding_mask: the mask for the src keys per batch (optional).

        Shape:
            see the docs in Transformer class.
        """
        src = self.norm1(src)
        src2 = self.self_attn(src, src, src, attn_mask=src_mask, key_padding_mask=src_key_padding_mask)[0]
        src = src + self.dropout1(src2)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(self.norm2(src)))))
        src = src + self.dropout2(src2)
        return src


class TransformerEncoderIntermediate(nn.TransformerEncoder):
    def __self__(self, encoder_layer: nn.TransformerEncoderLayer, num_layers: int):
        super().__init__(encoder_layer, num_layers)
        super().forward

    def forward(
        self, src: Tensor, mask: Optional[Tensor] = None, src_key_padding_mask: Optional[Tensor] = None
    ) -> List[Tensor]:
        output = src
        outputs = []

        for mod in self.layers:
            output = mod(output, src_mask=mask, src_key_padding_mask=src_key_padding_mask)
            outputs.append(output)

        if self.norm is not None:
            output = self.norm(output)
            outputs[-1] = output

        return outputs


class AcousticModel(nn.Module, metaclass=ABCMeta):
    _feature_size: int
    _output_size: int
    _upscale_factor: float
    _d_model: int

    @property
    def feature_size(self) -> int:
        return self._feature_size

    @property
    def output_size(self) -> int:
        return self._output_size

    @property
    def d_model(self) -> int:
        return self._d_model

    @property
    def upscale_factor(self) -> float:
        return self._upscale_factor

    @abstractmethod
    def downsampled_lengths(self, lengths: Tensor) -> Tensor: ...

    @abstractmethod
    def forward(self, batch: Batch, _predict: bool = False) -> Tuple[List[Tensor], Tensor]: ...


class TransformerAcousticModel(AcousticModel):
    def __init__(
        self,
        frontend: Frontend,
        transformer: TransformerEncoderIntermediate,
        sequential_frontend: Optional[SequentialFrontend] = None,
        input_dropout_rate: float = 0,
        use_positional_embeddings: bool = True,
        elementwise_affine: bool = False,
    ) -> None:
        super().__init__()
        self._input_dropout = nn.Dropout(input_dropout_rate)
        self._frontend = frontend
        self._transformer = transformer
        self._feature_size = frontend.output_dimensions
        self._final_layer_norm = nn.LayerNorm(
            self._feature_size if sequential_frontend is None else sequential_frontend.output_dimensions,
            elementwise_affine=elementwise_affine,
        )
        self._positional_embeddings = (
            SinusoidalPositionEmbeddings(
                self._feature_size if sequential_frontend is None else sequential_frontend.output_dimensions
            )
            if use_positional_embeddings
            else None
        )
        self._sequential_frontend = sequential_frontend
        self._upscale_factor = 1 if sequential_frontend is None else sequential_frontend.upscale_factor
        self._d_model = typing.cast(nn.Linear, transformer.layers[0].linear1).in_features
        self._output_size = typing.cast(nn.Linear, transformer.layers[-1].linear2).out_features

    @property
    def feature_size(self) -> int:
        return self._feature_size

    @property
    def upscale_factor(self) -> float:
        return self._upscale_factor

    def forward(self, batch: Batch, _predict: bool = False) -> Tuple[List[Tensor], Tensor]:
        frontend_output = self._frontend(batch)
        # Permutes from (N, F, L) to (L, N, F) for transformer (L = sequence, N = batch, F = features)
        frontend_features = frontend_output.audio_features.permute(2, 0, 1)
        model_inputs = self._input_dropout(frontend_features)

        if self._sequential_frontend is not None:
            frontend_output = self._sequential_frontend(
                Batch(model_inputs.permute(1, 2, 0), frontend_output.lengths, frontend_output.language_ids)
            )
            model_inputs = frontend_output.audio_features.permute(2, 0, 1)

        transformer_output = self._transformer(
            self._positional_embeddings(model_inputs) if self._positional_embeddings is not None else model_inputs,
            # Inverse mask since TransformerEncoderLayer expects `True` in padding locations
            src_key_padding_mask=utils.mask_sequence(frontend_output.lengths, inverse=True),
        )

        return [self._final_layer_norm(output) for output in transformer_output], frontend_output.lengths

    def downsampled_lengths(self, lengths: Tensor) -> Tensor:
        lengths = self._frontend.lengths(lengths)
        if self._sequential_frontend is None:
            return lengths
        return self._sequential_frontend.downsampled_lengths(lengths)

    @classmethod
    def from_config(cls, layer_config: TransformerAcousticModelConfig, feature_size: int):
        transformer_config = layer_config.transformer
        frontend = frontend_from_config(layer_config.frontend, feature_size, layer_config.elementwise_affine)
        previous_output_size = frontend.output_dimensions
        sequential_frontend_config = layer_config.sequential_frontend
        if sequential_frontend_config is not None:
            sequential_frontend = SequentialFrontend.from_config(sequential_frontend_config, previous_output_size)
            previous_output_size = sequential_frontend.output_dimensions
        else:
            sequential_frontend = None

        return cls(
            frontend,
            TransformerEncoderIntermediate(
                PreLMTransformerEncoderLayer(
                    previous_output_size,
                    transformer_config.heads,
                    transformer_config.feedforward_neurons,
                    transformer_config.dropout_rate,
                    transformer_config.activation,
                    layer_config.elementwise_affine,
                ),
                transformer_config.num_layers,
            ),
            sequential_frontend,
            transformer_config.dropout_rate,
            transformer_config.positional_embeddings,
            layer_config.elementwise_affine,
        )


@torch.jit.script
def zero_mean_unit_var_norm(features: Tensor, lengths: Tensor, mask: Tensor) -> Tensor:
    means = (features.sum(1) / lengths).unsqueeze(1)
    deviations = (features - means) * mask
    variances = (deviations**2).sum(1) / lengths
    return ((features - means) / (variances.unsqueeze(1) + 1e-7).sqrt()) * mask


def _freeze_module(model: nn.Module, trainable: bool = False) -> None:
    for parameter in model.parameters():
        parameter.requires_grad = trainable


class Wav2Vec2AcousticModel(AcousticModel):
    def __init__(
        self,
        model_id: str,
        sampling_rate: int = 16_000,
        freeze_feature_encoder: bool = True,
        freeze_feature_projection: bool = False,
        freeze_encoder: bool = False,
        load_pretrained_weights: bool = True,
        maximum_encoder_layers: int | None = None,
    ) -> None:
        super().__init__()
        self._feature_options, _ = Wav2Vec2FeatureExtractor.get_feature_extractor_dict(model_id)
        expected_sampling_rate = self._feature_options["sampling_rate"]
        if sampling_rate != expected_sampling_rate:
            raise ValueError(
                "Audio resampling config and the sampling rate required by Wav2Vec2 do not match. "
                f"Expected {expected_sampling_rate}kHz, got {sampling_rate}kHz"
            )

        if load_pretrained_weights:
            self._model = typing.cast(Wav2Vec2Model, Wav2Vec2Model.from_pretrained(model_id))
        else:
            self._model = Wav2Vec2Model(typing.cast(Wav2Vec2Config, Wav2Vec2Config.from_pretrained(model_id)))

        # Slice encoder layers if requested to save memory and compute
        if maximum_encoder_layers is not None:
            self._model.encoder._layers = self._model.encoder.layers[:maximum_encoder_layers]

        self._model.train(self.training)
        # Freeze the feature encoder as in https://arxiv.org/abs/2109.11680
        if freeze_feature_encoder:
            self._model.freeze_feature_encoder()
        # Freeze feature projection and encoder manually since there is no API method for it
        if freeze_feature_projection:
            _freeze_module(self._model.feature_projection)
        if freeze_encoder:
            _freeze_module(self._model.encoder)

        self._use_attention_mask = self._feature_options["return_attention_mask"]
        self._normalize = self._feature_options["do_normalize"]
        self._feature_size = self._feature_options["feature_size"]
        self._upscale_factor = 1
        config = self._model.config
        self._d_model = config.hidden_size
        # Falls back to d_model if no adapter with a different hidden size is used
        self._output_size = config.output_hidden_size or self._d_model
        self._sampling_rate = sampling_rate
        self._length_functions = [
            frontend.conv_length(kernel_size, stride, use_padding=False)
            for kernel_size, stride in zip(config.conv_kernel, config.conv_stride)
        ]

    @property
    def model(self) -> Wav2Vec2Model:
        return self._model

    def downsampled_lengths(self, lengths: Tensor) -> Tensor:
        for convolution_function in self._length_functions:
            lengths = convolution_function(lengths)
        return lengths

    def forward(self, batch: Batch, _predict: bool = False) -> Tuple[List[Tensor], Tensor]:
        sequence_mask = utils.mask_sequence(batch.lengths)
        output_hidden_states = self._model(
            (
                zero_mean_unit_var_norm(batch.audio_features, batch.lengths, sequence_mask)
                if self._normalize
                else batch.audio_features
            ),
            sequence_mask.long() if self._use_attention_mask else None,
            output_hidden_states=True,
        ).hidden_states

        # Transpose all hidden states to time dimension first for CTC
        return (
            [hidden_state.transpose(0, 1) for hidden_state in output_hidden_states],
            self.downsampled_lengths(batch.lengths),
        )


@dataclass
class UnfreezeSchedule:
    def __init__(
        self,
        feature_extractor: int | None = None,
        feature_projection: int | None = None,
        encoder_steps_remaining: int | None = None,
    ):
        self._steps = 0
        self._steps_remaining = [
            feature_extractor,
            feature_projection,
            encoder_steps_remaining,
        ]

    def step(self, acoustic_model: AcousticModel):
        if not isinstance(acoustic_model, Wav2Vec2AcousticModel):
            raise ValueError(
                f"Found an unsupported acoustic module type while updating an unfreeze schedule: {type(acoustic_model)}"
            )

        for index, layer in enumerate(
            (
                acoustic_model._model.feature_extractor,
                acoustic_model._model.feature_projection,
                acoustic_model._model.encoder,
            )
        ):
            # Skip layers that are already unfrozen or aren't scheduled to be
            steps = self._steps_remaining[index]
            if steps is None:
                continue

            # Take a step
            steps -= 1
            # Disable schedule and unfreeze the layer if all remaining steps are complete
            if steps <= 0:
                steps = None
                _freeze_module(layer, trainable=True)

            # Update remaining steps
            self._steps_remaining[index] = steps

    @classmethod
    def from_config(cls, config: UnfreezeScheduleConfig):
        return UnfreezeSchedule(
            config.feature_encoder_steps,
            config.feature_projection_steps,
            config.encoder_steps,
        )


@dataclass
class Predictions:
    """
    Contains predictions of a list of logits for each task and the number of output frames
    for each input utterance in a batch

    :param outputs: A dictionary of named output logit or probabilitiy tensor batches for each task
    :param lengths: A batch of the number of output frames
    """

    outputs: Dict[str, Tensor]
    lengths: Tensor

    def __len__(self) -> int:
        # Number of elements per batch
        return len(self.lengths)

    def task_count(self) -> int:
        return len(self.outputs)


AllophantCls = TypeVar("AllophantCls", bound="Allophant")


def _highest_specific_output_layer(graph: AttributeGraph) -> int | None:
    output_layer_indices = []
    for node in graph:
        for dependency in node.dependencies:
            output_match = ProjectionEntryConfig.OUTPUT_PATTERN.match(dependency)
            if output_match is not None and (layer_index := output_match.group(1)) is not None:
                output_layer_indices.append(int(layer_index))

    # Get exclusive index of the highest specific output layer or None, if only the final output is included
    return max(output_layer_indices) + 1 if output_layer_indices else None


class Allophant(nn.Module):
    def __init__(
        self,
        acoustic_model: AcousticModel,
        attribute_graph: AttributeGraph,
        blank_offset: int,
        projection_config: ProjectionConfig,
        attribute_indexer: PhoneticAttributeIndexer | None = None,
    ):
        super().__init__()
        self._acoustic_model = acoustic_model
        if attribute_indexer is not None and projection_config.phoneme_layer != PhonemeLayerType.SHARED:
            language_allophones = attribute_indexer.language_allophones
        else:
            language_allophones = None

        self._projection = HierarchicalProjection(
            acoustic_model.output_size,
            attribute_graph,
            blank_offset,
            projection_config.dependency_blanks,
            language_allophones,
            attribute_indexer,
            projection_config.acoustic_model_dropout,
            projection_config.embedding_composition,
        )
        self._classes = list(attribute_graph.names())

    @property
    def acoustic_model(self) -> AcousticModel:
        return self._acoustic_model

    @property
    def d_model(self) -> int:
        return self._acoustic_model.d_model

    @property
    def feature_size(self) -> int:
        return self._acoustic_model.feature_size

    @property
    def classes(self) -> List[str]:
        return self._classes

    @classmethod
    def from_config(
        cls: Type[AllophantCls],
        architecture: Architecture,
        feature_size: int,
        sample_rate: int,
        attribute_graph: AttributeGraph,
        attribute_indexer: PhoneticAttributeIndexer | None = None,
        load_pretrained_weights: bool = True,
    ) -> AllophantCls:
        layer_config = architecture.acoustic_model
        match layer_config:
            case TransformerAcousticModelConfig():
                acoustic_model = TransformerAcousticModel.from_config(layer_config, feature_size)
            case Wav2Vec2PretrainedConfig(model_id, freeze_feature_encoder, freeze_feature_projection, freeze_encoder):
                acoustic_model = Wav2Vec2AcousticModel(
                    model_id,
                    sample_rate,
                    freeze_feature_encoder,
                    freeze_feature_projection,
                    freeze_encoder,
                    load_pretrained_weights,
                    # Limit the pre-trained wav2vec2 model to at most the
                    # highest specific output layer required by the configuration for efficiency
                    _highest_specific_output_layer(attribute_graph),
                )
            case Wav2Vec2Model():
                raise NotImplementedError("Training Wav2Vec2 from scratch is not yet implemented")
            case model:
                raise ValueError(f"Unsupported model type: {type(model)}")

        return cls(
            acoustic_model,
            attribute_graph,
            architecture.loss.BLANK_OFFSET,
            architecture.projection,
            attribute_indexer,
        )

    def forward(self, batch: Batch, target_feature_indices: Tensor | None = None, predict: bool = False) -> Predictions:
        outputs, lengths = self._acoustic_model(batch, predict)
        return Predictions(
            self._projection(
                outputs,
                lengths,
                batch.language_ids,
                target_feature_indices,
                predict,
            ),
            lengths,
        )

    def map_allophones(self, phone_logits: Tensor, language_ids: Tensor) -> Tensor:
        return self._projection.map_allophones(phone_logits, language_ids)

    @property
    def upscale_factor(self) -> float:
        return self._acoustic_model.upscale_factor

    @property
    def projection(self) -> HierarchicalProjection:
        return self._projection

    def log_probabilities(self, outputs: Tensor) -> Tensor:
        return functional.log_softmax(outputs, -1)

    def downsampled_lengths(self, lengths: Tensor) -> Tensor:
        return self._acoustic_model.downsampled_lengths(lengths)

    def l2_penalty(self) -> Tensor | None:
        """
        Computes the l2 penalty for the allophone layer to be added to the loss

        :return: The l2 penalty for the allophone layer given the current
            weights or `None` if no allophone layer is present in the architecture
        """
        return self._projection.l2_penalty()



from collections.abc import Iterable, Iterator
from dataclasses import dataclass
from importlib.metadata import version
from io import TextIOWrapper
from os import path
import typing
from typing import Any, BinaryIO, ClassVar, Dict, Generic, List, Tuple, Type, TypeVar
import gzip

import torch
from torch import Tensor
from torchaudio.models import decoder
from torchaudio.models.decoder import CTCDecoder, CTCHypothesis

from marshmallow import Schema
from mashumaro.types import SerializationStrategy
from mashumaro.mixins.json import DataClassJSONMixin
import marshmallow_dataclass

import allophant
from allophant import phonemes
from allophant.phonemes import Action
from allophant import utils
from allophant.config import FeatureSet
from allophant.phonetic_features import ArticulatoryAttributes, PhonemeIndexer, PhoneticIndexerState
from allophant.utils import MarshmallowDataclassLoadMixin, PathOrFile, PathOrFileBinary


# Major, Minor, Patch
CURRENT_FORMAT_VERSION = (1, 1, 0)
SUPPORTED_VERSIONS = [CURRENT_FORMAT_VERSION]


@marshmallow_dataclass.add_schema
@dataclass
class PredictionMetaData(MarshmallowDataclassLoadMixin):
    Schema: ClassVar[Type[Schema]]

    prediction_arguments: str
    corpus_type: str
    languages: List[str]
    feature_set: FeatureSet
    indexer_state: PhoneticIndexerState
    classifiers: List[str]
    label_inventories: Dict[str, List[str]] | None = None
    package_version: str = version(allophant.__package__)
    format_version: Tuple[int, int, int] = CURRENT_FORMAT_VERSION


@dataclass
class UtterancePrediction(DataClassJSONMixin):
    language: str
    utterance_id: str
    predictions: Dict[str, List[List[str]]]
    labels: List[List[str]] | None = None


def levensthein_substitutions(expected: List[str], actual: List[str]) -> List[Tuple[Action, str, str]]:
    return phonemes.to_substitutions(expected, actual, phonemes.levensthein_operations(expected, actual)[0])


class ActionSerializationStrategy(SerializationStrategy, use_annotations=True):
    def serialize(self, value: Action) -> int:
        return int(value)

    def deserialize(self, value: int) -> Action:
        return Action.from_int(value)


@dataclass
class UtteranceEdits(DataClassJSONMixin):
    language: str
    utterance_id: str
    expected: Dict[str, List[str]]
    edit_operations: Dict[str, List[Tuple[Action, str, str]]]

    class Config:
        serialization_strategy = {
            Action: ActionSerializationStrategy(),
        }


def _infer_gzip(filepath: PathOrFile) -> bool:
    """
    Infers gzip format from a path or named file object if it points to a file with the `.gz` extension

    :param filepath: A path or named file object to infer the format from

    :return: `True` if the path or file object path points to a file with the `.gz` extension
    """
    return path.splitext(utils.get_filepath(filepath))[1] == ".gz"


T = TypeVar("T")


class JsonlReader(Generic[T]):
    def __init__(self, file: PathOrFileBinary, gzip: bool | None = None) -> None:
        """
        Creates a prediction reader for a path or file object to read predictions from

        :param file: A file or path object of the predictions either
            uncompressed or gzip compressed.
        :param gzip: Whether to treat the input file as gzip compressed.
            If the `gzip` argument was not given or `None`, compression is inferred if the path ends in ".gz"
        """
        self._wrapped_file = file
        self._gzip = _infer_gzip(file) if gzip is None else gzip

    def read_meta(self) -> Any:
        return None

    def process_line(self, line) -> T:
        return line

    def __iter__(self) -> Iterator[T]:
        for line in self._file:
            yield self.process_line(line)

    def __enter__(self):
        self._file = TextIOWrapper(
            typing.cast(
                BinaryIO,
                gzip.open(self._wrapped_file, "r") if self._gzip else utils.file_from(self._wrapped_file, "rb"),
            ),
            encoding="utf-8",
        )
        self._metadata = self.read_meta()
        return self

    def __exit__(self, *_) -> None:
        self._file.close()


class PredictionReader(JsonlReader[UtterancePrediction]):
    def read_meta(self) -> PredictionMetaData:
        return PredictionMetaData.loads(self._file.readline())

    @property
    def metadata(self) -> PredictionMetaData:
        return self._metadata

    def process_line(self, line) -> UtterancePrediction:
        return UtterancePrediction.from_json(line)


class StatisticsReader(JsonlReader[UtteranceEdits]):
    def read_meta(self) -> PredictionMetaData:
        return PredictionMetaData.loads(self._file.readline())

    @property
    def metadata(self) -> PredictionMetaData:
        return self._metadata

    def process_line(self, line) -> UtteranceEdits:
        return UtteranceEdits.from_json(line)


class JsonlWriter:
    def __init__(self, file: PathOrFileBinary, metadata: PredictionMetaData, gzip: bool = False) -> None:
        """
        Creates a prediction writer for a path or file object to save predictions to

        :param file: A file or path object to write metadata and predictions to
        :param gzip: Whether to compress outputs with gzip
        """
        self._wrapped_file = file
        self._gzip = _infer_gzip(file) if gzip is None else gzip
        self._meta_data = metadata

    def __enter__(self):
        self._file = TextIOWrapper(
            typing.cast(
                BinaryIO,
                gzip.open(self._wrapped_file, "x") if self._gzip else utils.file_from(self._wrapped_file, "xb"),
            ),
            encoding="utf-8",
        )
        self._file.write(self._meta_data.dumps() + "\n")
        return self

    def __exit__(self, *_) -> None:
        self._file.close()

    def write(self, serialized: DataClassJSONMixin) -> None:
        self._file.write(str(serialized.to_json()) + "\n")


class GreedyCTCDecoder:
    def __init__(self, blank_index: int = 0):
        super().__init__()
        self._blank_index = blank_index

    def __call__(self, log_emissions: Tensor, lengths: Tensor) -> List[List[CTCHypothesis]]:
        batch_max = torch.max(log_emissions, dim=-1)
        outputs = []
        for i, indices in enumerate(batch_max.indices):
            length = lengths[i]
            indices = indices[:length]
            decoded, sizes = torch.unique_consecutive(indices, return_counts=True)
            non_blanks = decoded != self._blank_index
            # Compute flashlight CTCDecoder compatible 1-based timesteps
            timesteps = (sizes.cumsum(0) - sizes + 1)[non_blanks]

            # Compute score as the sum of the highest probabilities as in the flashlight CTCDecoder
            outputs.append([CTCHypothesis(decoded[non_blanks], [], batch_max.values[i, :length].sum(), timesteps)])
        return outputs


class BeamCTCDecoder:
    def __init__(self, tokens: List[str], beam_width: int, n_best: int = 1, blank_index: int = 0) -> None:
        blank_token = tokens[blank_index]
        self._decoder = decoder.ctc_decoder(
            lexicon=None,
            tokens=tokens,
            blank_token=blank_token,
            # Ignore silence
            sil_token=blank_token,
            beam_size=beam_width,
            nbest=n_best,
            log_add=True,
        )

    def __call__(self, log_emissions: Tensor, lengths: Tensor | None = None) -> Any:
        # Compatibility with CTCDecoder, which requires regular probabilities
        return self._decoder(log_emissions.exp(), lengths)


def _ctc_decoder(categories: Iterable[str], beam_width: int = 1, n_best: int = 1) -> CTCDecoder:
    assert n_best <= beam_width, "N-best can not exceed beam width"
    # Optimized decoder for greedy decoding with log probabilities
    if beam_width == 1:
        return GreedyCTCDecoder()

    return BeamCTCDecoder(["<blank>", *categories], beam_width, n_best)


class FeatureDecoder:
    def __init__(self, feature_attributes: ArticulatoryAttributes, beam_width: int = 1, n_best: int = 1):
        self.attributes = feature_attributes
        self.feature_matrix = feature_attributes.dense_feature_table.long()
        self.decoder = _ctc_decoder(feature_attributes.phonemes, beam_width, n_best)


def feature_decoders(
    indexer: PhonemeIndexer,
    beam_width: int = 1,
    feature_names: Iterable[str] | None = None,
    n_best: int = 1,
) -> Dict[str, CTCDecoder]:
    return {
        name: _ctc_decoder(indexer.feature_categories(name), beam_width, n_best)
        for name in (indexer.feature_names if feature_names is None else feature_names)
    }



#!/usr/bin/env python
from collections.abc import Sequence
from argparse import ArgumentParser, FileType
import sys
from typing import Dict, List, Tuple

import numpy as np

from allophant.evaluation import MultilingualEvaluationResults


def _print_single_category_error_rates(
    title: str, results: List[Tuple[str, Dict[str, float]]], category_error_rates: Sequence[float]
) -> None:
    print(title)

    for (language, _), error_rates in zip(results, category_error_rates):
        print(language, error_rates, sep=",")

    print("Average", sum(category_error_rates) / len(category_error_rates), sep=",")


def main(args: Sequence[str] | None = None) -> None:
    if args is None:
        args = sys.argv[1:]

    parser = ArgumentParser()
    parser.add_argument(
        "results_file", type=FileType("r", encoding="utf-8"), help="Path to a results file from the evaluation command"
    )
    parser.add_argument(
        "-l",
        "--languages",
        type=lambda codes: set(codes.split(",")),
        help="Comma separated list of language codes to display the results and averages for",
    )

    arguments = parser.parse_args(args)

    language_subset = arguments.languages

    results_data = MultilingualEvaluationResults.load(arguments.results_file)
    # Don't include the total as a language in the report
    del results_data.results["total"]
    results = [
        (language, language_results.error_rates)
        for language, language_results in results_data.results.items()
        if language_subset is None or language in language_subset
    ]

    if language_subset is not None and len(language_subset) > len(results):
        raise ValueError(
            f"Languages from -l/--languages are missing from the results file: [{', '.join(language_subset - {language for language, _ in results})}]"
        )

    print("Evaluation Arguments", results_data.evaluation_arguments)
    print("Package Version", results_data.package_version)

    first_result = results[0][1]

    phone = [] if "phone" in first_result else None
    phoneme = [] if "phoneme" in first_result else None
    attribute = []

    for _, error_rates in results:
        if phone is not None:
            phone.append(error_rates.pop("phone") * 100)
        if phoneme is not None:
            phoneme.append(error_rates.pop("phoneme") * 100)
        attribute.append(list(error_rates.values()))

    if phone is not None:
        _print_single_category_error_rates("phoneme", results, phone)

    if phoneme is not None:
        _print_single_category_error_rates("phoneme", results, phoneme)

    attribute_error_rates = np.array(attribute) * 100
    if not attribute_error_rates.size:
        print("No Attribute Error Rates")
        return

    for (language, _), error_rates in zip(results, attribute_error_rates.mean(1)):
        print(language, error_rates, sep=",")

    for attribute, error_rates in zip(results[0][1], attribute_error_rates.mean(0)):
        print(attribute, error_rates, sep=",")

    print("Average", attribute_error_rates.mean(), sep=",")


if __name__ == "__main__":
    main(sys.argv[1:])



from collections.abc import Iterable, Iterator, Sequence
import itertools
import logging
from logging import FileHandler, Filter, LogRecord, Logger
from typing import List, TypeVar
import unicodedata
import re

from regex import regex

from allophant.phonemes import IpaSegmenter
from allophant.utils import AnyPath


TONES = [
    "꜀",
    "꜁",
    "꜂",
    "꜃",
    "꜄",
    "꜅",
    "꜆",
    "꜇",
    "꜈",
    "꜉",
    "꜊",
    "꜋",
    "꜌",
    "꜍",
    "꜎",
    "꜏",
    "꜐",
    "꜑",
    "꜒",
    "꜓",
    "꜔",
    "꜕",
    "꜖",
    "ꜗ",
    "ꜘ",
    "ꜙ",
    "ꜚ",
    "ꜛ",
    "ꜜ",
    "ꜝ",
    "ꜞ",
    "ꜟ",
]


TONE_PATTERN = f"[{''.join(TONES)}]"


def filter_tones(inventory: Iterable[str]) -> Iterator[str]:
    for phoneme in inventory:
        if not re.search(phoneme, TONE_PATTERN):
            yield phoneme


VOWELS = {
    "a",
    "e",
    "i",
    "o",
    "u",
    "y",
    "æ",
    "ø",
    "œ",
    "ɐ",
    "ɑ",
    "ɒ",
    "ɔ",
    "ɘ",
    "ə",
    "ɚ",
    "ɛ",
    "ɜ",
    "ɝ",
    "ɞ",
    "ɤ",
    "ɨ",
    "ɪ",
    "ɯ",
    "ɵ",
    "ɶ",
    "ʉ",
    "ʊ",
    "ʌ",
    "ʏ",
}


class _MissingPhonemeFilter(Filter):
    _DEFAULT_LANGUAGE = "unknown"

    def __init__(self):
        self._seen = set()
        self.language = self._DEFAULT_LANGUAGE

    def filter(self, record: LogRecord) -> bool:
        phoneme = record.msg
        entry = (self.language, phoneme)
        if entry in self._seen:
            return False
        self._seen.add(entry)
        record.msg = f"Missing phoneme segment: ({self.language}) {phoneme!r}"
        return True

    def reset(self) -> None:
        self._seen = set()
        self.language = self._DEFAULT_LANGUAGE


class SegmentationLogger:
    def __init__(self) -> None:
        self._reset_steps = 0
        self._segmentation_logger = logging.getLogger("allophant.phoneme_segmentation")
        self._segmentation_logger.setLevel(logging.WARNING)
        self._filter = _MissingPhonemeFilter()
        self._segmentation_logger.addFilter(self._filter)

    def add_file_handler(self, log_path: AnyPath) -> None:
        self._segmentation_logger.addHandler(FileHandler(log_path, mode="w", encoding="utf-8"))

    def context_language(self, language: str) -> None:
        self._filter.language = language

    @property
    def log(self) -> Logger:
        return self._segmentation_logger

    def reset(self) -> None:
        if self._reset_steps > 0:
            self._reset_steps -= 1
        else:
            self._filter.reset()

    def suppress_resets(self, count: int = 1) -> None:
        self._reset_steps += count


SEGMENTATION_LOGGER = SegmentationLogger()


I = TypeVar("I", bound=int | Sequence[int])


class SegmentationProcessor:
    def pre_process(self, phoneme: str, _phoneme_iterator: Iterator[str]) -> str:
        return phoneme

    def post_process(self, sub_segments: List[str], _phoneme_iterator: Iterator[str]) -> Sequence[str]:
        return sub_segments


class IpaSentenceSegmenter:
    def __init__(self, dictionary: List[str], processor: SegmentationProcessor | None = None) -> None:
        self._segmenter = IpaSegmenter(dictionary)
        self._processor = SegmentationProcessor() if processor is None else processor

    @property
    def word_segmenter(self) -> IpaSegmenter:
        return self._segmenter

    def __call__(self, phonetic_sentences: Iterable[List[str]]) -> Iterator[List[str]]:
        return (self._segmenter.segment_words_checked(sentence) for sentence in phonetic_sentences)

    def lossy_segment(self, phonetic_sentences: Iterable[List[str]]) -> Iterator[List[str]]:
        for sentence in phonetic_sentences:
            sentence_phonemes = []
            phoneme_iterator = iter(sentence)
            for phoneme in phoneme_iterator:
                pre_processed = self._processor.pre_process(phoneme, phoneme_iterator)
                sub_segments = self._processor.post_process(
                    self._segmenter.segment(pre_processed),
                    phoneme_iterator,
                )
                sentence_phonemes.extend(sub_segments)
                if len(sub_segments) != 1:
                    if "".join(sub_segments) != pre_processed:
                        SEGMENTATION_LOGGER.log.warning(pre_processed + " (Missing sub-segment when split)")
                    else:
                        SEGMENTATION_LOGGER.log.warning(pre_processed)

            yield sentence_phonemes


def _is_mark(character: str) -> bool:
    category = unicodedata.category(character)
    return category.endswith("m") or category == "Sk" or category.startswith("M")


def is_vowel_only_segment(segment: str) -> bool:
    return all(character in VOWELS for character in segment if not _is_mark(character))


def base_phonemes(segment: str) -> Iterator[str]:
    # Any symbols that are not marks or diacritics.
    # Should leave only phoneme base characters
    return (character for character in segment if not _is_mark(character))


def complex_with_vowel(base_phonemes: Iterable[str]) -> bool:
    count = 0
    has_vowel = False
    for character in base_phonemes:
        if character in VOWELS:
            has_vowel = True
        count += 1

    # Any segment with more than one base phoneme and a vowel
    return has_vowel and count > 1


def is_multi_vowel(base_phonemes: Iterable[str]) -> bool:
    count = 0
    for character in base_phonemes:
        # Not a multi vowel segment if there's a symbol that is not a
        # diacritic, mark or a vowel
        if character not in VOWELS:
            return False
        count += 1

    # Only multi-vowel if there is more than one
    return count > 1


def split_complex_segment(segment: str) -> List[str]:
    vowels = []
    prefix = ""
    # Get all grapheme clusters - including vowels with attached combining characters
    for grapheme in regex.finditer(r"\X", segment):
        # Skip IPA marks that count as separate
        grapheme = grapheme.group()
        if len(grapheme) == 1 and _is_mark(grapheme):
            if not vowels:
                prefix += grapheme
            else:
                vowels[len(vowels) - 1] += grapheme
        else:
            vowels.append(prefix + grapheme)
            prefix = ""

    # Handle combining mark only segments, this mostly affects tone contours, keeping them together
    if prefix:
        vowels.append(prefix)

    return vowels


def split_all_complex_segments(segments: Iterable[str]) -> Iterator[str]:
    return itertools.chain.from_iterable(map(split_complex_segment, segments))


def split_phoneme_segment(segment: str) -> List[List[str]]:
    return [split_complex_segment(subsegment) for subsegment in segment.split("|")]



from collections import defaultdict
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
import itertools
from os import path
from typing import Any, ClassVar, Dict, Generic, Iterator, List, Literal, Protocol, Tuple, TypeVar, runtime_checkable
from pandas.io.formats.format import math

from torch import Tensor
import torchaudio
from mutagen import File
from torchaudio.transforms import Resample

from allophant.datasets.phonemes import PhonemeTranscription, PhonemeSource
from allophant.phonetic_features import LanguageInventories
from allophant import utils
from allophant.utils import AnyPath


@dataclass
class LanguageInfo:
    """
    ISO 6393 language code and phoneme inventory for a language

    :param code: ISO 6393 language code
    :param phoneme_inventory: A list of IPA segments that form that languages phoneme inventory
    :param phoneme_mappings: Inventory mappings from Grapheme to Phoneme engines used to produce transcriptions
    """

    code: str

    phoneme_inventory: List[str]
    phoneme_mappings: List[Dict[str, List[str]]]


class TranscribedUtterance(Protocol):
    """A transcribed utterance with a unique ID"""

    utterance_id: str


@runtime_checkable
class PhoneticallyTranscribedUtterance(TranscribedUtterance, Protocol):
    """An utterance with a phonemic transcription words and the original orthographic forms for each word"""

    phonemes: PhonemeTranscription


@runtime_checkable
class PhoneticallySegmentedUtterance(TranscribedUtterance, Protocol):
    "An utterance with a phonemic transcription"
    phonemes: List[str]


T = TypeVar("T", bound=TranscribedUtterance)
I = TypeVar("I", bound=LanguageInfo)


@dataclass
class LanguageData(Generic[I, T]):
    """
    Metadata and transcribed utterances for a monolingual corpus

    :param info: Language-specific metadata
    :param transcribed_samples: A list of transcribed utterances for a language
    """

    info: I
    transcribed_samples: List[T]

    def __getitem__(self, index: int) -> T:
        return self.transcribed_samples[index]

    def __len__(self) -> int:
        return len(self.transcribed_samples)


@dataclass
class _LanguageEntry(Generic[I]):
    """
    Monolingual subset range in a `MultilingualCorpus`

    :param info: Language-specific metadata
    :param offset: Start offset of the monolingual subset in the multilingual utterance list
    :param num_utterances: The number of utterances available for monolingual subset
    """

    info: I
    offset: int
    num_utterances: int


@dataclass
class IndexedEntry(Generic[T]):
    """
    Represents an entry in a `MultilingualCorpus` with a language ID

    :param language_id: Language ID within the `MultilingualCorpus` that the entry belongs to
    :param entry: An entry in the `MultilingualCorpus` - usually an utterance
    """

    language_id: int
    entry: T


@dataclass
class AudioInfo:
    """
    Container for audio metadata

    :param sample_rate: The sample rate
    :param bits_per_sample: The bit depth
    :param num_channels: The number of audio channels
    """

    sample_rate: int = 16_000
    bits_per_sample: int = 16
    num_channels: int = 1

    @classmethod
    def none(cls):
        """Creates a fallback instance with every parameter set to `0`"""
        return cls(0, 0, 0)


class MultilingualCorpus(Generic[I, T]):
    """
    Generic interface for a multilingual corpus that allows both access to monolingual subsets and random access to utterances across languages
    """

    UTTERANCE_TYPE = TranscribedUtterance
    _SAMPLE_RATES = [44100]

    def __init__(
        self,
        base_directory: AnyPath,
        language_data: Iterable[LanguageData[I, T]],
        audio_subdirectory: str,
        audio_extension: str,
        limits: Dict[str, int] | int | None = None,
        resample: int | None = None,
        phoneme_sources: Sequence[PhonemeSource] | None = None,
        meta_data: Dict[str, Any] | None = None,
    ) -> None:
        """
        Creates a new multilingual corpus that adheres to the format of a base
        directory with a single subdirectory containing all audio files with
        the same extension. For construction, an iterable has to be provided which
        yields `LanguageData` instances for all or a subset of the languages
        contained in the corpus.

        :param base_directory: Base directory of the multilingual corpus
        :param language_data: An iterable yielding `LanguageData` instances which
            contain metadata and a list of utterances for monolingual subsets from the corpus.
            If multiple `LanguageData` instances are provided, a `ValueError` will be raised.
        :param audio_subdirectory: Base directory of the multilingual corpus
        :param audio_extension: File extension of the audio files contained in the dataset
        :param limits: Either a map of language to integer to set an utterance
            limit per language or a single integer to set an utterance limit globally
        :param resample: If given, all utterances will be resampled to the given sample rate when they are loaded
        :param phoneme_sources: List of sources for phoneme transcriptions available in this corpus. By default,
            it is assumed that there is one phoneme source which was manual transcribed.
            This corresponds to :py:attr:`~allophant.dataset.phonemes.PhonemeTranscriptionType.MANUAL`
        :param meta_data: An optional dictionary of key value pairs
        """
        self._meta_data = meta_data
        # Absolutize data path for later re-use from another location
        self._base_directory = path.abspath(path.expanduser(base_directory))
        self._audio_extension = audio_extension
        self._audio_subdirectory = audio_subdirectory
        self._phoneme_sources = [PhonemeSource.MANUAL] if phoneme_sources is None else phoneme_sources
        self._languages: Dict[str, _LanguageEntry] = {}
        self._language_ids: Dict[str, int] = {}
        self._language_list: List[str] = []
        self._utterances: List[IndexedEntry[T]] = []
        self._num_utterances = 0

        if resample is not None:
            self._resamplers = {sample_rate: Resample(sample_rate, resample) for sample_rate in self._SAMPLE_RATES}
        else:
            self._resamplers = None

        for i, language in enumerate(language_data):
            language_code = language.info.code
            limit = utils.global_or_local_limit(limits, language_code)
            num_utterances = len(language.transcribed_samples)
            if limit is not None:
                num_utterances = min(num_utterances, limit)

            if language_code in self._language_ids:
                raise ValueError(f"Duplicate language data for code: {language_code}")
            self._languages[language_code] = _LanguageEntry(language.info, self._num_utterances, num_utterances)
            self._language_ids[language_code] = i
            self._language_list.append(language_code)

            # Add utterances for the current language, optionally limited
            self._utterances.extend(
                IndexedEntry(i, transcription)
                for transcription, _ in zip(language.transcribed_samples, utils.limit_indices(limit))
            )
            self._num_utterances += num_utterances

        # Assumes audio sampling rate (if no resampling is used), channels and bit depth are same for every audio file in the corpus
        self._audio_info = AudioInfo.none()
        if resample is not None:
            self._audio_info.sample_rate = resample

    @property
    def meta_data(self) -> Dict[str, Any] | None:
        return self._meta_data

    @property
    def audio_info(self) -> AudioInfo:
        """
        Returns audio information about the `MultilingualCorpus`.
        Currently, only the sample rate is returned and only if resampling is enabled,
        since the original sample rates may differ between audio files
        """
        return self._audio_info

    @property
    def phoneme_sources(self) -> Sequence[PhonemeSource]:
        return self._phoneme_sources

    @staticmethod
    def empty() -> "MultilingualCorpus":
        """Creates a dummy empty corpus"""
        return MultilingualCorpus("", [], "", "")

    def path_from_utterance(self, language: str, utterance_id: str) -> str:
        """
        Retrieves the path to an utterance with the given language and ID from the corpus

        :param language: language code of the utterance in the format used by the corpus
        :param utterance_id: The utterance_id in the format used by the corpus

        :return: The path to the requested utterance
        """
        return path.join(
            self._base_directory, language, self._audio_subdirectory, f"{utterance_id}.{self._audio_extension}"
        )

    def path_for(self, indexed_transcription: IndexedEntry[T]) -> str:
        """
        Retrieves the path to an utterance for the given indexed transcription from this corpus

        :param indexed_transcription: An `IndexedEntry` from the corpus

        :return: The path to the requested utterance
        """
        return self.path_from_utterance(
            self._language_list[indexed_transcription.language_id], indexed_transcription.entry.utterance_id
        )

    def path(self, index: int) -> str:
        """
        Retrieves the path to an utterance for the given utterance index

        :param index: Index of an utterance in the multilingual corpus

        :return: The path to the requested utterance
        """
        return self.path_for(self._utterances[index])

    def audio_from_utterance(self, language: str, utterance_id: str) -> Tuple[Tensor, int]:
        """
        Retrieves the audio and original sample rate of an utterance with the given language and ID from the corpus

        :param language: language code of the utterance in the format used by the corpus
        :param utterance_id: The utterance_id in the format used by the corpus

        :return: A tuple of (possibly resampled) audio samples as a tensor and the original sample rate before resampling
        """
        samples, sample_rate = torchaudio.load(self.path_from_utterance(language, utterance_id))  # pyright: ignore
        return (samples if self._resamplers is None else self._resamplers[sample_rate](samples), sample_rate)

    def audio_for(self, indexed_transcription: IndexedEntry[T]) -> Tuple[Tensor, int]:
        """
        Retrieves the audio and original sample rate of an utterance for the given indexed transcription from this corpus

        :param indexed_transcription: An `IndexedEntry` from the corpus

        :return: A tuple of (possibly resampled) audio samples as a tensor and the original sample rate before resampling
        """
        return self.audio_from_utterance(
            self._language_list[indexed_transcription.language_id], indexed_transcription.entry.utterance_id
        )

    def audio(self, index: int) -> Tuple[Tensor, int]:
        """
        Retrieves the audio and original sample rate of an utterance for the given utterance index

        :param index: Index of an utterance in the multilingual corpus

        :return: A tuple of (possibly resampled) audio samples as a tensor and the original sample rate before resampling
        """
        return self.audio_for(self._utterances[index])

    @property
    def languages(self) -> List[str]:
        return self._language_list

    def utterance_languages(self) -> Iterator[str]:
        """
        Generates language codes for each utterance in the corpus

        :return: Language codes for each utterance in the corpus
        """
        # Repeat language codes for all languages in order
        # - only works since dict is guaranteed to be ordered in newer Python versions
        for code, language_entry in self._languages.items():
            for _ in range(language_entry.num_utterances):
                yield code

    def monolingual_index_range(self, language_code: str) -> range:
        """
        Retrieves the range from the start to the stop index of utterance in
        the given language

        :param language_code: An ISO 6393 language code

        :return: The utterance index range for `language_code`
        """
        language_entry = self._languages[language_code]
        offset = language_entry.offset
        return range(offset, offset + language_entry.num_utterances)

    def _monolingual_islice(self, language_entry: _LanguageEntry) -> Iterator[IndexedEntry[T]]:
        offset = language_entry.offset
        return itertools.islice(self._utterances, offset, offset + language_entry.num_utterances)

    def monolingual_subset(self, language_code: str) -> LanguageData[I, T]:
        """
        Retrieves the monolingual subcorpus with the given language code

        :param language_code: An ISO 6393 language code

        :return: The monolingual subcorpus for `language_code`
        """
        language_entry = self._languages[language_code]
        return LanguageData(
            language_entry.info, [utterance.entry for utterance in self._monolingual_islice(language_entry)]
        )

    def subset(self, language_codes: Iterable[str]) -> "MultilingualCorpus[I, T]":
        """
        Creates a multilingual subcorpus from the given language codes

        :param language_codes: An iterable of ISO 6393 language codes

        :return: A subcorpus containing only the languages included in `language_codes`
        """
        return MultilingualCorpus(
            self._base_directory,
            (self.monolingual_subset(language_code) for language_code in language_codes),
            self._audio_subdirectory,
            self._audio_extension,
        )

    def shared_inventory(self) -> List[str]:
        """
        Collects the shared phoneme inventory from all languages in the corpus.
        The resulting IPA segments are returned in sorted order.

        :return: The shared phoneme inventory of the corpus in sorted order
        """
        # Sorted for more determinism
        return sorted(
            {
                phoneme
                for language_entry in self._languages.values()
                for phoneme in language_entry.info.phoneme_inventory
            }
        )

    def language_id_inventories(self) -> LanguageInventories:
        """
        Creates a mapping between corpus-specific language IDs and phoneme
        inventories, including the language codes for each ID

        :return: A mapping from language IDs to phoneme inventories and a list of language codes corresponding to the IDs
        """
        return LanguageInventories(
            {
                self.language_id(language_entry.info.code): language_entry.info.phoneme_inventory
                for language_entry in self._languages.values()
            },
            self.languages,
        )

    def language_id(self, language: str) -> int:
        """
        Retrieves the corpus-specific language ID for a language

        :param language: An ISO 6393 language code

        :return: The ID of the given language in the corpus
        """
        return self._language_ids[language]

    def language(self, language_id: int) -> str:
        """
        Retrieves the ISO 6393 language code for a corpus-specific language ID

        :param language_id: A language ID from this corpus

        :return: The ISO6393 language code corresponding to the `language_id` in this corpus
        """
        return self._language_list[language_id]

    def inventory(self, language: str) -> List[str]:
        """
        Retrieves the phoneme inventory for a language

        :param language: An ISO 6393 language code

        :return: The phoneme inventory for the given language
        """
        return self._languages[language].info.phoneme_inventory

    def _language_lengths(self, language_entry: _LanguageEntry) -> Iterator[int]:
        """
        Calculates the number of samples after resampling for every utterance in the `language_entry`

        :param language_entry: A monolingual subset for which to compute sample counts

        :return: A generator of sample counts for each utterance in `language_entry`
        """
        sample_rate = self._audio_info.sample_rate
        offset = language_entry.offset
        for utterance_path in self._monolingual_islice(language_entry):
            audio_path = self.path_for(utterance_path)
            audio_file = File(audio_path)
            if audio_file is None:
                raise ValueError(f"Could not guess audio file type for {audio_path!r}")
            file_info = audio_file.info
            # Get original sample count
            original_sample_count = file_info.length * file_info.sample_rate
            # Get target samples after resampling - adapted from _apply_sinc_resample_kernel
            # in https://github.com/pytorch/audio/blob/7ac3e2e237e443baf91dfbf9893fca114c1c6001/torchaudio/functional/functional.py
            yield int(math.ceil((sample_rate * original_sample_count) / file_info.sample_rate))

    def read_lengths(self) -> Iterator[Tuple[str, Iterator[int]]]:
        """
        Calculates the number of samples after resampling for every utterance in the corpus one language at a time

        :return: A generator of ISO 6393 language code and sample count generator tuples
        """
        return ((code, self._language_lengths(language_entry)) for code, language_entry in self._languages.items())

    def __getitem__(self, index: int) -> IndexedEntry[T]:
        return self._utterances[index]

    def __iter__(self) -> Iterator[LanguageData[I, T]]:
        for language in self._language_list:
            yield self.monolingual_subset(language)

    def __len__(self) -> int:
        return self._num_utterances

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._base_directory!r}, {self._languages!r})"


@dataclass
class SplitMetaData:
    train: Dict[str, Any] | None
    dev: Dict[str, Any] | None
    test: Dict[str, Any] | None


C = TypeVar("C", bound=MultilingualCorpus)


@dataclass
class MultilingualSplits(Generic[C]):
    """
    Represents train, dev and test splits of a corpus

    :param train: A subset used for training
    :param dev: A subset used for validation
    :param test: A held-out test set
    """

    SPLIT_NAMES: ClassVar[Tuple[Literal["train"], Literal["dev"], Literal["test"]]] = ("train", "dev", "test")

    train: C
    dev: C
    test: C

    def audio_info(self) -> AudioInfo:
        """
        Retrieves the `AudioInfo` for the first non-empty split or :func:`AudioInfo.none` if all splits are empty.

        :return: The `AudioInfo` for the first non-empty split
        """
        no_info = AudioInfo.none()
        return next(
            (split.audio_info for split in (self.train, self.dev, self.test) if split.audio_info != no_info), no_info
        )

    def meta_data(self) -> SplitMetaData:
        return SplitMetaData(self.train.meta_data, self.dev.meta_data, self.test.meta_data)

    @classmethod
    def single(cls, corpus: C, split: Literal["train"] | Literal["dev"] | Literal["test"] = "test"):
        """
        Creates `MultilingualSplits` with only a single split

        :param corpus: The `MultilingualCorpus` to be used as a split
        :param split: The type of split the `corpus` should be assigned to

        :return: `MultilingualSplits` with a single split where `corpus` is assigned to `split`
        """
        splits = {split: corpus}
        corpus_type = corpus.__class__
        for data_split in ("train", "dev", "test"):
            if data_split not in splits:
                splits[data_split] = corpus_type.empty()  # type: ignore

        return cls(**splits)

    def __iter__(self) -> Iterator[C]:
        # Yields splits sequentially
        yield self.train
        yield self.dev
        yield self.test
